
C***********************************************************************
C   Portions of Models-3/CMAQ software were developed or based on      *
C   information from various groups: Federal Government employees,     *
C   contractors working on a United States Government contract, and    *
C   non-Federal sources (including research institutions).  These      *
C   research institutions have given the Government permission to      *
C   use, prepare derivative works, and distribute copies of their      *
C   work in Models-3/CMAQ to the public and to permit others to do     *
C   so.  EPA therefore grants similar permissions for use of the       *
C   Models-3/CMAQ software, but users are requested to provide copies  *
C   of derivative works to the Government without restrictions as to   *
C   use by others.  Users are responsible for acquiring their own      *
C   copies of commercial software associated with Models-3/CMAQ and    *
C   for complying with vendor requirements.  Software copyrights by    *
C   the MCNC Environmental Modeling Center are used with their         *
C   permissions subject to the above restrictions.                     *
C***********************************************************************

C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      SUBROUTINE GET_OBS_FG(LGRID, JDATE, JTIME, F)

C-----------------------------------------------------------------------
C Function:
C   Get the function value and gradient of the observation operator at the
C   current values of the control variables

C Revision History:
C   Apr 2007 Kumaresh Singh and Adrian Sandu at Virginia Tech: created
C   Jun 2009 Peter Percell at UH-IMAQS: rewrote
C-----------------------------------------------------------------------

      USE GRID_CONF             ! horizontal & vertical domain specifications
      USE FOURDVAR_MOD          ! 4DVAR data
      USE ADJOINT_FILES         ! adjoint file data and functions

      IMPLICIT NONE

C Include Files:
      INCLUDE SUBST_IOPARMS     ! I/O parameters definitions
      INCLUDE SUBST_FILES_ID    ! file name parameters

C Parameters:
      REAL(8), PARAMETER :: UNC_OBS = 0.08D0, ADD_ERR = 1.0D-3

C Arguments:
      REAL, POINTER :: LGRID(:, :, :, :)
      INTEGER, INTENT(IN) :: JDATE ! current model date, coded YYYYDDD
      INTEGER, INTENT(IN) :: JTIME ! current model time, coded HHMMSS
      REAL(8), INTENT(INOUT) :: F ! cumulative value of cost function (for local domain)

C External Functions:
      INTEGER, EXTERNAL :: JUNIT

C Local Variables:
      CHARACTER(16) :: PNAME = 'GET_OBS_FG'
      CHARACTER(96) :: XMSG = ' '

      INTEGER, SAVE :: LDATE = 9999999 ! last date observation were read (YYYYDDD)

      INTEGER, SAVE :: N_OBS_RECORDS ! number of observation records

      INTEGER, ALLOCATABLE, SAVE :: OBS_HOUR(:) ! observation times (HH, 0-23 UTC)
      INTEGER, ALLOCATABLE, SAVE :: OBS_COL(:) ! observation column indices
      INTEGER, ALLOCATABLE, SAVE :: OBS_ROW(:) ! observation row indices
      INTEGER, ALLOCATABLE, SAVE :: OBS_LAY(:) ! observation layer indices
      REAL, ALLOCATABLE, SAVE :: OBS_CONC(:, :) ! observed concentrations (ppmV)

      INTEGER :: ALLOCSTAT, OSTAT

      CHARACTER(512) :: OBS_FILE ! full path to observation input file
      INTEGER :: IUNIT           ! FORTRAN unit number for observation input file

      INTEGER :: IJUNK

      REAL :: OBS_CHK(1:NCOLS, 1:NROWS, 1:NLAYS, 1:N_OBS_SPC) ! observed species checkpoint

      INTEGER :: K              ! loop counter

      INTEGER :: HH             ! current hour
      INTEGER :: C, R, L, S, OS ! column, row, layer and species indices

      REAL(8) :: STD            ! standard deviation

      REAL(8) :: D

C-----------------------------------------------------------------------

      IF ( N_OBS_SPC /= 1 ) THEN
         XMSG = 'Assume just one observed species (for now)'
         CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT3)
      END IF

      IF ( JDATE /= LDATE ) THEN
         LDATE = JDATE

         IUNIT = JUNIT()
         IF ( IUNIT < 0 ) THEN
            XMSG = 'No available unit for IO'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
         END IF

         CALL NAMEVAL(OBS_DATA, OBS_FILE)
         OPEN(UNIT=IUNIT, FILE=OBS_FILE, STATUS='OLD', ACTION='READ', IOSTAT=OSTAT)
         IF ( OSTAT /= 0 ) THEN
            XMSG = 'Failure opening observation input file ' //
     &           TRIM(OBS_FILE)
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
         END IF

         READ(IUNIT, *) N_OBS_RECORDS, IJUNK
         READ(IUNIT, *)         ! skip next 3 lines
         READ(IUNIT, *)
         READ(IUNIT, *)

         IF ( ALLOCATED(OBS_HOUR) ) DEALLOCATE(OBS_HOUR)
         IF ( ALLOCATED(OBS_COL) ) DEALLOCATE(OBS_COL)
         IF ( ALLOCATED(OBS_ROW) ) DEALLOCATE(OBS_ROW)
         IF ( ALLOCATED(OBS_LAY) ) DEALLOCATE(OBS_LAY)
         IF ( ALLOCATED(OBS_CONC) ) DEALLOCATE(OBS_CONC)

         ALLOCATE(
     &        OBS_HOUR(1:N_OBS_RECORDS),
     &        OBS_COL(1:N_OBS_RECORDS),
     &        OBS_ROW(1:N_OBS_RECORDS),
     &        OBS_LAY(1:N_OBS_RECORDS),
     &        OBS_CONC(1:N_OBS_RECORDS, 1:N_OBS_SPC),
     &        STAT = ALLOCSTAT)
         IF ( ALLOCSTAT /= 0 ) THEN
            XMSG = 'Failure allocating ' //
     &           'OBS_HOUR, OBS_COL, OBS_ROW, OBS_LAY, OBS_CONC'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
         END IF

C Read observed concentrations
         DO K = 1, N_OBS_RECORDS
             READ(IUNIT, *)
     &           OBS_HOUR(K), OBS_COL(K), OBS_ROW(K), OBS_LAY(K), OBS_CONC(K, 1)
             OBS_CONC(K, 1) = OBS_CONC(K, 1) / 1000.0 ! ppbV => ppmV
         END DO

         CLOSE(UNIT=IUNIT, IOSTAT=OSTAT)
         IF ( OSTAT /= 0 ) THEN
            XMSG = 'Failure closing observation input file ' //
     &           TRIM(OBS_FILE)
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
         END IF
      END IF                    ! JDATE /= LDATE

C Read current model calculated values
      CALL RD_ADJ_FILE(OBS_CHK, JDATE, JTIME, ADJ_OBS_CHK)

C Update the cost function F and gradient LGRID
      HH = JTIME / 10000
      OS = 1
      S = OBS_TO_CGRID_MAP(OS)
      DO K = 1, N_OBS_RECORDS
         IF ( OBS_HOUR(K) > HH ) EXIT

         IF ( OBS_HOUR(K) == HH ) THEN
            C = OBS_COL(K) - COLSX_PE(1,MYPE+1) + 1
            R = OBS_ROW(K) - ROWSX_PE(1,MYPE+1) + 1
            IF ( C >= 1 .AND. C <= NCOLS .AND.
     &           R >= 1 .AND. R <= NROWS ) THEN
               L = OBS_LAY(K)

               STD = UNC_OBS * OBS_CONC(K, OS) + ADD_ERR
               D = (OBS_CHK(C, R, L, OS) - OBS_CONC(K, OS)) / STD
               F = F + D**2 / 2.0D0
               LGRID(C, R, L, S) = LGRID(C, R, L, S) + D / STD
            END IF
         END IF
      END DO

      END SUBROUTINE GET_OBS_FG
