
C***********************************************************************
C   Portions of Models-3/CMAQ software were developed or based on      *
C   information from various groups: Federal Government employees,     *
C   contractors working on a United States Government contract, and    *
C   non-Federal sources (including research institutions).  These      *
C   research institutions have given the Government permission to      *
C   use, prepare derivative works, and distribute copies of their      *
C   work in Models-3/CMAQ to the public and to permit others to do     *
C   so.  EPA therefore grants similar permissions for use of the       *
C   Models-3/CMAQ software, but users are requested to provide copies  *
C   of derivative works to the Government without restrictions as to   *
C   use by others.  Users are responsible for acquiring their own      *
C   copies of commercial software associated with Models-3/CMAQ and    *
C   for complying with vendor requirements.  Software copyrights by    *
C   the MCNC Environmental Modeling Center are used with their         *
C   permissions subject to the above restrictions.                     *
C***********************************************************************

C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      PROGRAM DRIVER

C-----------------------------------------------------------------------
C Function:
C   Driver for sensitivity analysis

C Revision History:
C   Apr 2007 Kumaresh Singh and Adrian Sandu at Virginia Tech: created
C   Aug 2007 Peter Percell at UH-IMAQS: rewrote
C-----------------------------------------------------------------------

      USE PCGRID_DEFN           ! inherits HGRD_DEFN and CGRID_SPCS
      USE VGRD_DEFN             ! vertical layer specifications
      USE CGRID_SPCS            ! CGRID species number and offsets
      USE FOURDVAR_MOD          ! 4DVAR data
      USE ADJOINT_FILES         ! adjoint file data and functions

      IMPLICIT NONE

C Include Files:
      INCLUDE SUBST_IOPARMS     ! I/O parameters definitions
      INCLUDE SUBST_IODECL      ! I/O definitions and declarations
      INCLUDE SUBST_FILES_ID    ! file name parameters
#ifdef parallel
      INCLUDE SUBST_MPICH       ! MPI definitions and parameters
#endif

C Load the mechanism COMMON ... (next two INCLUDE files)
      INCLUDE SUBST_RXCMMN      ! chemical mechamism reactions COMMON
      INCLUDE SUBST_RXDATA      ! chemical mechamism reactions DATA

C Parameters:

C External Functions (not declared in IODECL3.EXT):
      INTEGER, EXTERNAL :: ENVINT
      LOGICAL, EXTERNAL :: ENVYN, BWRITE3

C Local Variables:
      CHARACTER(16) :: PNAME = 'DRIVER'
      CHARACTER(96) :: XMSG = ' '

      REAL :: CLOCK             ! Wall-clock time (sec) at initialization.
      INTEGER :: PAR_ERR        ! Error code from parallel initialization.

      INTEGER :: SDATE          ! start date, format YYYYDDD
      INTEGER :: STIME          ! start time, format HHMMSS
      INTEGER :: TSTEP(2)       ! time step vector (HHMMSS)
                                ! TSTEP(1) = local output step
                                ! TSTEP(2) = sciproc sync. step (chem)
      INTEGER :: NSTEPS         ! run duration: number of output time steps
      INTEGER :: NREPS          ! number of sync steps per output step

      INTEGER :: JDATE          ! current model date, coded YYYYDDD
      INTEGER :: JTIME          ! current model time, coded HHMMSS

      REAL, POINTER :: CGRID(:, :, :, :)

      REAL, POINTER :: LGRID(:, :, :, :) ! gradient of cost function

      REAL, POINTER :: BASE_IC(:, :, :, :) ! base initial conditions

      LOGICAL :: WR_LGRID
      LOGICAL :: WR_LGRID_BC_SF

      INTEGER :: DAY_NUM        ! day number

      LOGICAL :: XFIRST

      INTEGER :: S              ! species index

      INTEGER :: MX_FG          ! maximum number of function + gradient evaluations
      CHARACTER(80) :: VARDESC  ! environment variable description
      INTEGER :: STATUS         ! ENVINT, ENVYN status

C L-BFGS-B Variables (see routines.f for descriptions):
      CHARACTER(60) :: TASK, CSAVE
      LOGICAL :: LSAVE(4)
      INTEGER :: N, M, IPRINT, ISAVE(44)
      INTEGER, ALLOCATABLE :: NBD(:), IWA(:)
      REAL(8) :: F, FACTR, PGTOL, DSAVE(29)
      REAL(8), ALLOCATABLE :: L(:), U(:), WA(:), X(:, :, :), G(:, :, :)
      REAL, ALLOCATABLE :: XS(:, :, :), GS(:, :, :)

      INTEGER :: ALLOCSTAT

      REAL(8) :: GL_F           ! global value of cost function

#ifdef parallel
      INTEGER :: IERROR         ! MPI error code
#endif

      INTERFACE
         SUBROUTINE INITSCEN(CGRID, SDATE, STIME, TSTEP, NSTEPS)
            IMPLICIT NONE
            REAL, POINTER        :: CGRID(:, :, :, :)
            INTEGER, INTENT(OUT) :: SDATE, STIME
            INTEGER, INTENT(OUT) :: TSTEP(2)
            INTEGER, INTENT(OUT) :: NSTEPS
         END SUBROUTINE INITSCEN
         SUBROUTINE DRIVER_FWD(CGRID, JDATE, JTIME, TSTEP, NSTEPS, NREPS,
     &     DAY_NUM, XFIRST, OBS_FILE)
            IMPLICIT NONE
            REAL, POINTER          :: CGRID(:, :, :, :)
            INTEGER, INTENT(INOUT) :: JDATE, JTIME
            INTEGER, INTENT(IN)    :: TSTEP(2), NSTEPS
            INTEGER, INTENT(IN)    :: NREPS
            INTEGER, INTENT(OUT)   :: DAY_NUM
            LOGICAL, INTENT(INOUT) :: XFIRST
            CHARACTER(16), INTENT(IN), OPTIONAL :: OBS_FILE
         END SUBROUTINE DRIVER_FWD
         SUBROUTINE DRIVER_BWD(LGRID, JDATE, JTIME, TSTEP, NSTEPS, NREPS,
     &     DAY_NUM, XFIRST, F)
            IMPLICIT NONE
            REAL, POINTER          :: LGRID(:, :, :, :)
            INTEGER, INTENT(INOUT) :: JDATE, JTIME
            INTEGER, INTENT(IN)    :: TSTEP(2), NSTEPS
            INTEGER, INTENT(IN)    :: NREPS
            INTEGER, INTENT(INOUT) :: DAY_NUM
            LOGICAL, INTENT(INOUT) :: XFIRST
            REAL(8), INTENT(INOUT) :: F
         END SUBROUTINE DRIVER_BWD
         SUBROUTINE GATHER_BNDY_3D(L_ARRAY, G_ARRAY, JDATE, JTIME)
            IMPLICIT NONE
            REAL, INTENT(IN) :: L_ARRAY(:, :, :)
            REAL, INTENT(OUT) :: G_ARRAY(:, :, :)
            INTEGER, INTENT(IN) :: JDATE, JTIME
         END SUBROUTINE GATHER_BNDY_3D
         SUBROUTINE SCATTER_BNDY_3D(G_ARRAY, L_ARRAY, JDATE, JTIME)
            IMPLICIT NONE
            REAL, INTENT(IN) :: G_ARRAY(:, :, :)
            REAL, INTENT(OUT) :: L_ARRAY(:, :, :)
            INTEGER, INTENT(IN) :: JDATE, JTIME
         END SUBROUTINE SCATTER_BNDY_3D
      END INTERFACE

C-----------------------------------------------------------------------

C Get number of species, and starting indices for CGRID array.
      CALL CGRID_MAP(NSPCSD, GC_STRT, AE_STRT, NR_STRT, TR_STRT)

C Start up processor communication and define horizontal domain decomposition
C and vertical layer structure
      CALL PAR_INIT(PNAME, NSPCSD, CLOCK, PAR_ERR)
      IF ( PAR_ERR /= 0 ) THEN
         XMSG = 'Error in PAR_INIT'
         CALL M3EXIT(PNAME, 0, 0, XMSG, XSTAT2)
      END IF

C Initialize PCGRID
      IF ( .NOT. PCGRID_INIT() ) THEN
         XMSG = 'Failure defining horizontal domain'
         CALL M3EXIT(PNAME, 0, 0, XMSG, XSTAT2)
      END IF

      CGRID => PCGRID(1:NCOLS, 1:NROWS, :, :)

C Initialize conc field: Copy IC's to CONC file as step 0
C Convention: the input file concentration units are always ppmV.
      CALL INITSCEN(CGRID, SDATE, STIME, TSTEP, NSTEPS)

C Verify input file header consistency with COORD.EXT and run duration
      CALL FLCHECK(SDATE, STIME, TSTEP(1))

C Initialize 4DVAR data
      CALL FOURDVAR_INIT(SDATE, STIME)

      ALLOCATE(
     &     LGRID(1:NCOLS, 1:NROWS, 1:NLAYS, 1:NSPCSD),
     &     BASE_IC(1:NCOLS, 1:NROWS, 1:NLAYS, 1:NSPCSD),
     &     BC_SF(1:NBNDY, 1:NLAYS, 1:N_CNTL_SPC),
     &     LGRID_BC_SF(1:NBNDY, 1:NLAYS, 1:N_CNTL_SPC),
     &     STAT = ALLOCSTAT)
      IF ( ALLOCSTAT /= 0 ) THEN
         XMSG = 'Failure allocating LGRID, BASE_IC, BC_SF, LGRID_BC_SF'
         CALL M3EXIT(PNAME, SDATE, STIME, XMSG, XSTAT1)
      END IF

C Set the current date and time to the start date and time
      JDATE = SDATE
      JTIME = STIME

C Get the synchronization step, TSTEP(2), and number of repetitions per output
C step, NREPS
C (Note: TSTEP(2) and NREPS are constant for a run because checkpoints are
C written every sync step and the IOAPI requires a constant output time step.)
      CALL SYNCSTEP(JDATE, JTIME, TSTEP, NREPS)

C Create new adjoint files and open for update
      CALL OP_ADJ_FILE(JDATE, JTIME, TSTEP, ADJ_CHEM_CHK)
      CALL OP_ADJ_FILE(JDATE, JTIME, TSTEP, ADJ_OBS_CHK)

      WR_LGRID =
     &     ENVYN('WR_LGRID', 'Write LGRID output', .FALSE., STATUS)
      IF ( WR_LGRID )
     &     CALL OP_ADJ_FILE(JDATE, JTIME, TSTEP, ADJ_LGRID)

      CALL OP_ADJ_FILE(JDATE, JTIME, TSTEP, ADJ_BC_SF)

      WR_LGRID_BC_SF =
     &     ENVYN('WR_LGRID_BC_SF', 'Write LGRID_BC_SF output', .FALSE., STATUS)
      IF ( WR_LGRID_BC_SF )
     &     CALL OP_ADJ_FILE(JDATE, JTIME, TSTEP, ADJ_LGRID_BC_SF)

C Save the base initial conditions
      BASE_IC(:, :, :, :) = CGRID(:, :, :, :)

C Get the maximum number of function + gradient evaluations allowed
      MX_FG = 10                ! default
      VARDESC = 'Maximum number of function + gradient evaluations allowed'
      MX_FG = ENVINT('ADJ_MX_FG', VARDESC, MX_FG, STATUS)
      IF ( STATUS /= 0 .OR. MX_FG < 0 ) THEN
         XMSG = VARDESC
         CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT2)
      END IF

C===============================================
C Set up input to the L-BFGS-B minimization code
C===============================================

C Specify the dimension N of the problem
      N = GL_NBNDY * NLAYS * N_CNTL_SPC

C Allocate memory for XS & GS on all processors
      ALLOCATE(
     &     XS(GL_NBNDY, NLAYS, N_CNTL_SPC),
     &     GS(GL_NBNDY, NLAYS, N_CNTL_SPC),
     &     STAT = ALLOCSTAT)
      IF ( ALLOCSTAT /= 0 ) THEN
         XMSG = 'Failure allocating XS, GS'
         CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
      END IF

      IF ( MYPE == 0 ) THEN

C Request output every iteration
         IPRINT = 1

C Specify the tolerances in the stopping criteria
         FACTR = 1.0D+7
         PGTOL = 1.0D-5

C Specify the number M of limited memory corrections stored
         M = 5

C Allocate memory for L-BFGS-B workspace arrays
         ALLOCATE(
     &        NBD(N),
     &        IWA(3 * N),
     &        L(N),
     &        U(N),
     &        X(GL_NBNDY, NLAYS, N_CNTL_SPC),
     &        G(GL_NBNDY, NLAYS, N_CNTL_SPC),
     &        WA((2 * M + 4) * N + 12 * M * (M + 1)),
     &        STAT = ALLOCSTAT)
         IF ( ALLOCSTAT /= 0 ) THEN
            XMSG = 'Failure allocating L-BFGS-B arrays'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
         END IF

C Set NBD which defines the type of bounds imposed on the variables:
C L   specifies the lower bounds,
C U   specifies the upper bounds (which can remain undefined for this problem).
         NBD(1:N) = 1           ! lower bounds only
         L(1:N) = 0.0D0

C Start the iteration by initializing TASK
         TASK = 'START'

      END IF                    ! MYPE == 0

C Set the starting values of the control variables (boundary condition scaling
C factors) to 1.0
      XS(:, :, :) = 1.0

C===============================
C Beginning of minimization loop
C===============================
      DO

         IF ( MYPE == 0 ) THEN

C Call the L-BFGS-B code
            X(:, :, :) = DBLE(XS(:, :, :))
            G(:, :, :) = DBLE(GS(:, :, :))
            CALL SETULB(N, M, X, L, U, NBD, GL_F, G, FACTR, PGTOL, WA, IWA,
     &           TASK, IPRINT, CSAVE, LSAVE, ISAVE, DSAVE)
            XS(:, :, :) = REAL(X(:, :, :))

            IF ( .NOT. BWRITE3(ADJ_BC_SF, ALLVAR3, JDATE, JTIME, XS) ) THEN
               XMSG = 'Could not write to ' // TRIM(ADJ_BC_SF)
               CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
            END IF

         END IF                 ! MYPE == 0

#ifdef parallel
C Broadcast TASK and ISAVE to all processors
         CALL MPI_BCAST(TASK, 60, MPI_CHARACTER, 0, MPI_COMM_WORLD, IERROR)
         CALL MPI_BCAST(ISAVE, 44, MPI_INTEGER, 0, MPI_COMM_WORLD, IERROR)
#endif

         IF ( TASK(1:2) == 'FG' ) THEN ! compute F and G at current X

C Scatter the global XS array to the local BC_SF arrays
            CALL SCATTER_BNDY_3D(XS, BC_SF, JDATE, JTIME)

C Restore CGRID to the base initial conditions
            CGRID(:, :, :, :) = BASE_IC(:, :, :, :)

C Do forward simulation and write checkpoint states
            XFIRST = .TRUE.
            CALL DRIVER_FWD(CGRID, JDATE, JTIME, TSTEP, NSTEPS, NREPS,
     &           DAY_NUM, XFIRST, ADJ_OBS_CHK)

C Open checkpoint files as read-only
            IF ( .NOT. OPEN3(ADJ_CHEM_CHK, FSREAD3, PNAME) ) THEN
               XMSG = 'Could not open ' // TRIM(ADJ_CHEM_CHK) //
     &              ' file as OLD:READ-ONLY'
               CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
            END IF
            IF ( .NOT. OPEN3(ADJ_OBS_CHK, FSREAD3, PNAME) ) THEN
               XMSG = 'Could not open ' // TRIM(ADJ_OBS_CHK) //
     &              ' file as OLD:READ-ONLY'
               CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
            END IF

C Do backward (adjoint) solution and accumulate the value of the cost function
C on the local domain
            F = 0.0D0
            CALL DRIVER_BWD(LGRID, JDATE, JTIME, TSTEP, NSTEPS, NREPS,
     &           DAY_NUM, XFIRST, F)

C If requested, write the LGRID file
            IF ( WR_LGRID ) CALL WR_ADJ_FILE(LGRID, JDATE, JTIME, ADJ_LGRID)

C Accumulate the value of the cost function on the global domain
#ifdef parallel
            CALL MPI_ALLREDUCE(F, GL_F, 1, MPI_REAL8, MPI_SUM,
     &           MPI_COMM_WORLD, IERROR)
#else
            GL_F = F
#endif

C Gather the local LGRID_BC_SF arrays into the global GS array
            GS(:, :, :) = 0.0
            CALL GATHER_BNDY_3D(LGRID_BC_SF, GS, JDATE, JTIME)

C If requested, write the LGRID_BC_SF file
            IF ( MYPE == 0 .AND. WR_LGRID_BC_SF ) THEN
               IF ( .NOT. BWRITE3(ADJ_LGRID_BC_SF, ALLVAR3, JDATE, JTIME, GS) ) THEN
                  XMSG = 'Could not write to ' // TRIM(ADJ_LGRID_BC_SF)
                  CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
               END IF
            END IF

         ELSE IF ( TASK(1:5) .EQ. 'NEW_X' ) THEN

            IF ( ISAVE(34) >= MX_FG )
     &           TASK = 'STOP: Total no. of F and G evaluations exceeds limit'

         ELSE

C If TASK is neither FG nor NEW_X, then terminate the minimization
            EXIT

         END IF

      END DO
C=========================
C End of minimization loop
C=========================

C Clean up communications
      CALL PAR_TERM(PNAME, JDATE, JTIME, CLOCK)

      END PROGRAM DRIVER
