
C***********************************************************************
C   Portions of Models-3/CMAQ software were developed or based on      *
C   information from various groups: Federal Government employees,     *
C   contractors working on a United States Government contract, and    *
C   non-Federal sources (including research institutions).  These      *
C   research institutions have given the Government permission to      *
C   use, prepare derivative works, and distribute copies of their      *
C   work in Models-3/CMAQ to the public and to permit others to do     *
C   so.  EPA therefore grants similar permissions for use of the       *
C   Models-3/CMAQ software, but users are requested to provide copies  *
C   of derivative works to the Government without restrictions as to   *
C   use by others.  Users are responsible for acquiring their own      *
C   copies of commercial software associated with Models-3/CMAQ and    *
C   for complying with vendor requirements.  Software copyrights by    *
C   the MCNC Environmental Modeling Center are used with their         *
C   permissions subject to the above restrictions.                     *
C***********************************************************************

C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      SUBROUTINE DRIVER_BWD(LGRID, JDATE, JTIME, TSTEP, NSTEPS, NREPS,
     &     DAY_NUM, XFIRST, F)

C-----------------------------------------------------------------------
C Function:
C   Do backward run to generate adjoint trajectory

C Revision History:
C   Apr 2007 Kumaresh Singh and Adrian Sandu at Virginia Tech: created
C   Aug 2007 Peter Percell at UH-IMAQS: rewrote
C-----------------------------------------------------------------------

      USE GRID_CONF             ! horizontal & vertical domain specifications
      USE FOURDVAR_MOD          ! 4DVAR data

      IMPLICIT NONE

C Include Files:
      INCLUDE SUBST_IOPARMS     ! I/O parameters definitions

C Parameters:

C Arguments:
      REAL, POINTER :: LGRID(:, :, :, :) ! gradient of cost function
      INTEGER, INTENT(INOUT) :: JDATE ! current model date, coded YYYYDDD
      INTEGER, INTENT(INOUT) :: JTIME ! current model time, coded HHMMSS
      INTEGER, INTENT(IN) :: TSTEP(2) ! time step vector (HHMMSS)
                                      !   TSTEP(1) = local output step
                                      !   TSTEP(2) = sciproc sync. step (chem)
      INTEGER, INTENT(IN) :: NSTEPS   ! run duration: number of output time steps
      INTEGER, INTENT(IN) :: NREPS    ! number of model time steps per output step
      INTEGER, INTENT(INOUT) :: DAY_NUM ! day number
      LOGICAL, INTENT(INOUT) :: XFIRST
      REAL(8), INTENT(INOUT) :: F ! cumulative value of cost function (for local domain)

C Local Variables:
      LOGICAL, SAVE :: FIRSTIME = .TRUE.

      CHARACTER(16) :: PNAME = 'DRIVER_BWD'
      CHARACTER(96) :: XMSG = ' '

      INTEGER, ALLOCATABLE, SAVE :: NADVSS(:) ! number of advection sub-steps per sync step

      INTEGER :: ALLOCSTAT

      INTEGER :: SDATE          ! start date, format YYYYDDD
      INTEGER :: STIME          ! start time, format HHMMSS

      INTEGER :: STEP_NUM       ! current output time step number
      INTEGER :: IREP           ! model step number within this output step

      INTERFACE
         SUBROUTINE ADVSTEP(JDATE, JTIME, TSTEP, NADVSS)
            IMPLICIT NONE
            INTEGER, INTENT(IN)  :: JDATE, JTIME
            INTEGER, INTENT(IN)  :: TSTEP(2)
            INTEGER, INTENT(OUT) :: NADVSS(:)
         END SUBROUTINE ADVSTEP
         SUBROUTINE GET_OBS_FG(LGRID, JDATE, JTIME, F)
            IMPLICIT NONE
            REAL, POINTER          :: LGRID(:, :, :, :)
            INTEGER, INTENT(IN)    :: JDATE, JTIME
            REAL(8), INTENT(INOUT) :: F
         END SUBROUTINE GET_OBS_FG
         SUBROUTINE GET_BG_FG(JDATE, JTIME, F)
            IMPLICIT NONE
            INTEGER, INTENT(IN)    :: JDATE, JTIME
            REAL(8), INTENT(INOUT) :: F
         END SUBROUTINE GET_BG_FG
         SUBROUTINE SCIPROC_ADJ(LGRID, JDATE, JTIME, TSTEP, NADVSS, XFIRST)
            IMPLICIT NONE
            REAL, POINTER          :: LGRID(:, :, :, :)
            INTEGER, INTENT(INOUT) :: JDATE, JTIME
            INTEGER, INTENT(IN)    :: TSTEP(2)
            INTEGER, INTENT(IN)    :: NADVSS(:)
            LOGICAL, INTENT(INOUT) :: XFIRST
         END SUBROUTINE SCIPROC_ADJ
      END INTERFACE

C-----------------------------------------------------------------------

      IF ( FIRSTIME ) THEN
         FIRSTIME = .FALSE.

         ALLOCATE(
     &        NADVSS(1:NLAYS),
     &        STAT = ALLOCSTAT)
         IF ( ALLOCSTAT /= 0 ) THEN
            XMSG = 'Failure allocating NADVSS'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
         END IF
      END IF                    ! FIRSTIME

      LGRID(:, :, :, :) = 0.0
      LGRID_BC_SF(:, :, :) = 0.0

C Get the function value and gradient of the observation operator
      CALL GET_OBS_FG(LGRID, JDATE, JTIME, F)

C Do backward (adjoint) solution and write sensitivities (step backwards from
C final time)
      DO STEP_NUM = NSTEPS, 1, -1 ! output time step loop
C Get number of advection sub-steps per sync step, NADVSS(L)
         SDATE = JDATE
         STIME = JTIME
         CALL NEXTIME(SDATE, STIME, -TSTEP(1))

         IF ( SDATE < JDATE ) THEN
            DAY_NUM = DAY_NUM - 1
            CALL SET_INPUT_ENV(SDATE, STIME, DAY_NUM)
         END IF

         CALL ADVSTEP(SDATE, STIME, TSTEP, NADVSS)

         DO IREP = NREPS, 1, -1
            CALL SCIPROC_ADJ(LGRID, JDATE, JTIME, TSTEP, NADVSS, XFIRST)

C Get the function value and gradient of the observation operator
            CALL GET_OBS_FG(LGRID, JDATE, JTIME, F)
         END DO
      END DO

C Get the function value and gradient of the background operator
      CALL GET_BG_FG(JDATE, JTIME, F)

      END SUBROUTINE DRIVER_BWD
