
C***********************************************************************
C   Portions of Models-3/CMAQ software were developed or based on      *
C   information from various groups: Federal Government employees,     *
C   contractors working on a United States Government contract, and    *
C   non-Federal sources (including research institutions).  These      *
C   research institutions have given the Government permission to      *
C   use, prepare derivative works, and distribute copies of their      *
C   work in Models-3/CMAQ to the public and to permit others to do     *
C   so.  EPA therefore grants similar permissions for use of the       *
C   Models-3/CMAQ software, but users are requested to provide copies  *
C   of derivative works to the Government without restrictions as to   *
C   use by others.  Users are responsible for acquiring their own      *
C   copies of commercial software associated with Models-3/CMAQ and    *
C   for complying with vendor requirements.  Software copyrights by    *
C   the MCNC Environmental Modeling Center are used with their         *
C   permissions subject to the above restrictions.                     *
C***********************************************************************

C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      SUBROUTINE GET_BG_FG(JDATE, JTIME, NSTEPS, F)

C-----------------------------------------------------------------------
C Function:
C   Get the function value and gradient of the background operator at the
C   initial time

C Revision History:
C   Apr 2007 Kumaresh Singh and Adrian Sandu at Virginia Tech: created
C   Jun 2009 Peter Percell at UH-IMAQS: rewrote
C-----------------------------------------------------------------------

      USE GRID_CONF             ! horizontal & vertical domain specifications
      USE FOURDVAR_MOD          ! 4DVAR data

      IMPLICIT NONE

C Include Files:
#ifdef parallel
      INCLUDE SUBST_MPICH       ! MPI definitions and parameters
#endif

C Parameters:
      REAL(8), PARAMETER :: BG_WGT = 12.0D0

C Arguments:
      INTEGER, INTENT(IN) :: JDATE ! current model date, coded YYYYDDD
      INTEGER, INTENT(IN) :: JTIME ! current model time, coded HHMMSS
      INTEGER, INTENT(IN) :: NSTEPS ! run duration: number of output time steps
      REAL(8), INTENT(INOUT) :: F ! cumulative value of cost function (for local domain)

C Local Variables:
      LOGICAL, SAVE :: FIRSTIME = .TRUE.

      CHARACTER(16) :: PNAME = 'GET_BG_FG'
      CHARACTER(96) :: XMSG = ' '

      LOGICAL, SAVE :: S_BND_PE, E_BND_PE, N_BND_PE, W_BND_PE

      INTEGER, SAVE :: SZ       ! local zero index for southern boundary
      INTEGER, SAVE :: EZ       ! local zero index for eastern boundary
      INTEGER, SAVE :: NZ       ! local zero index for northern boundary
      INTEGER, SAVE :: WZ       ! local zero index for western boundary

      INTEGER :: STEP_NUM       ! current output time step number

      REAL(8) :: STD(1:NLAYS, 1:N_CNTL_SPC, 0:NSTEPS) ! (global) standard deviation

      INTEGER :: C, R, L, S     ! column, row, layer and species indices

      REAL(8) :: D

#ifdef parallel
      REAL(8) :: STD_LD         ! (local) standard deviation
      INTEGER :: IERROR         ! MPI error code
#endif

C-----------------------------------------------------------------------

      IF ( FIRSTIME ) THEN
         FIRSTIME = .FALSE.

         S_BND_PE = ( ROWSX_PE(1, MYPE+1) == 1 )
         E_BND_PE = ( COLSX_PE(2, MYPE+1) == GL_NCOLS )
         N_BND_PE = ( ROWSX_PE(2, MYPE+1) == GL_NROWS )
         W_BND_PE = ( COLSX_PE(1, MYPE+1) == 1 )

         SZ = 0
         EZ = NCOLS + 1
         NZ = NCOLS + NROWS + 3
         WZ = 2 * NCOLS + NROWS + 4
      END IF                    ! FIRSTIME

C Estimate the (global) standard deviation of each background value over each
C layer and step
      DO STEP_NUM = 0, NSTEPS
         DO S = 1, N_CNTL_SPC
            DO L = 1, NLAYS
#ifdef parallel
               STD_LD = 0.0D0
               IF ( S_BND_PE )
     &              STD_LD = STD_LD +
     &              SUM(DBLE(BG_BC_TS(SZ+1:SZ+NCOLS, L, S, STEP_NUM)))
               IF ( E_BND_PE )
     &              STD_LD = STD_LD +
     &              SUM(DBLE(BG_BC_TS(EZ+1:EZ+NROWS, L, S, STEP_NUM)))
               IF ( N_BND_PE )
     &              STD_LD = STD_LD +
     &              SUM(DBLE(BG_BC_TS(NZ+1:NZ+NCOLS, L, S, STEP_NUM)))
               IF ( W_BND_PE )
     &              STD_LD = STD_LD +
     &              SUM(DBLE(BG_BC_TS(WZ+1:WZ+NROWS, L, S, STEP_NUM)))
               CALL MPI_ALLREDUCE(STD_LD, STD(L, S, STEP_NUM), 1, MPI_REAL8,
     &              MPI_SUM, MPI_COMM_WORLD, IERROR)
#else
               STD(L, S, STEP_NUM) = 0.0D0
               IF ( S_BND_PE )
     &              STD(L, S, STEP_NUM) = STD(L, S, STEP_NUM) +
     &              SUM(DBLE(BG_BC_TS(SZ+1:SZ+NCOLS, L, S, STEP_NUM)))
               IF ( E_BND_PE )
     &              STD(L, S, STEP_NUM) = STD(L, S, STEP_NUM) +
     &              SUM(DBLE(BG_BC_TS(EZ+1:EZ+NROWS, L, S, STEP_NUM)))
               IF ( N_BND_PE )
     &              STD(L, S, STEP_NUM) = STD(L, S, STEP_NUM) +
     &              SUM(DBLE(BG_BC_TS(NZ+1:NZ+NCOLS, L, S, STEP_NUM)))
               IF ( W_BND_PE )
     &              STD(L, S, STEP_NUM) = STD(L, S, STEP_NUM) +
     &              SUM(DBLE(BG_BC_TS(WZ+1:WZ+NROWS, L, S, STEP_NUM)))
#endif
               STD(L, S, STEP_NUM) =
     &              BG_WGT * STD(L, S, STEP_NUM) / (2*(GL_NCOLS + GL_NROWS))
               IF ( ABS(STD(L, S, STEP_NUM)) < 1.0D-30 )
     &              STD(L, S, STEP_NUM) = 1.0D0
            END DO
         END DO
      END DO

C Update the cost function F and gradient LGRID_BC_TS
      DO STEP_NUM = 0, NSTEPS
         DO S = 1, N_CNTL_SPC
            DO L = 1, NLAYS

C South boundary
               IF ( S_BND_PE ) THEN
                  DO C = 1, NCOLS
                     D = (CNTL_BC_TS(SZ+C, L, S, STEP_NUM)
     &                    - BG_BC_TS(SZ+C, L, S, STEP_NUM))
     &                    / STD(L, S, STEP_NUM)
                     F = F + D**2 / 2.0D0
                     LGRID_BC_TS(SZ+C, L, S) =
     &                    LGRID_BC_TS(SZ+C, L, S) + D / STD(L, S, STEP_NUM)
                  END DO
               END IF

C East boundary
               IF ( E_BND_PE ) THEN
                  DO R = 1, NROWS
                     D = (CNTL_BC_TS(EZ+R, L, S, STEP_NUM)
     &                    - BG_BC_TS(EZ+R, L, S, STEP_NUM))
     &                    / STD(L, S, STEP_NUM)
                     F = F + D**2 / 2.0D0
                     LGRID_BC_TS(EZ+R, L, S) =
     &                    LGRID_BC_TS(EZ+R, L, S) + D / STD(L, S, STEP_NUM)
                  END DO
               END IF

C North boundary
               IF ( N_BND_PE ) THEN
                  DO C = 1, NCOLS
                     D = (CNTL_BC_TS(NZ+C, L, S, STEP_NUM)
     &                    - BG_BC_TS(NZ+C, L, S, STEP_NUM))
     &                    / STD(L, S, STEP_NUM)
                     F = F + D**2 / 2.0D0
                     LGRID_BC_TS(NZ+C, L, S) =
     &                    LGRID_BC_TS(NZ+C, L, S) + D / STD(L, S, STEP_NUM)
                  END DO
               END IF

C West boundary
               IF ( W_BND_PE ) THEN
                  DO R = 1, NROWS
                     D = (CNTL_BC_TS(WZ+R, L, S, STEP_NUM)
     &                    - BG_BC_TS(WZ+R, L, S, STEP_NUM))
     &                    / STD(L, S, STEP_NUM)
                     F = F + D**2 / 2.0D0
                     LGRID_BC_TS(WZ+R, L, S) =
     &                    LGRID_BC_TS(WZ+R, L, S) + D / STD(L, S, STEP_NUM)
                  END DO
               END IF

            END DO
         END DO
      END DO

      END SUBROUTINE GET_BG_FG
