C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      SUBROUTINE INIT_BNDY_COND(SDATE, STIME, TSTEP, NSTEPS, XS)

C-----------------------------------------------------------------------
C Function:
C   - combine the boundary condition files for the different species
C     classes into one file
C   - concatenate the combined files in time if the run spans more than
C     one day
C   - initialize the boundary condition control variables
C   - save the initial values of the boundary condition control variables
C     as "background" values

C Revision History:
C   Aug 2009 by Peter Percell at UH-IMAQS: Created
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      USE GRID_CONF             ! horizontal & vertical domain specifications
      USE FOURDVAR_MOD          ! 4DVAR data
      USE ADJOINT_FILES         ! adjoint file data and functions

      IMPLICIT NONE

C Include Files:
      INCLUDE SUBST_GC_ADV      ! gas chem advection species and map table
      INCLUDE SUBST_AE_ADV      ! aerosol advection species and map table
      INCLUDE SUBST_NR_ADV      ! non-react advection species and map table
      INCLUDE SUBST_TR_ADV      ! tracer advection species and map table
      INCLUDE SUBST_IOPARMS     ! I/O parameters definitions
      INCLUDE SUBST_IODECL      ! I/O definitions and declarations
      INCLUDE SUBST_FILES_ID    ! file name parameters

C Arguments:
      INTEGER, INTENT(IN) :: SDATE ! start date, format YYYYDDD
      INTEGER, INTENT(IN) :: STIME ! start time, format HHMMSS
      INTEGER, INTENT(IN) :: TSTEP(2) ! time step vector (HHMMSS)
                                ! TSTEP(1) = local output step
                                ! TSTEP(2) = sciproc sync. step (chem)
      INTEGER, INTENT(IN) :: NSTEPS ! run duration, as number of output time steps
      REAL, INTENT(OUT)   :: XS(1:, 1:, 1:, 0:) ! boundary condition control variables

C External Functions (not declared in IODECL3.EXT):
      LOGICAL, EXTERNAL :: BWRITE3
      INTEGER, EXTERNAL :: NAME2FID

C Local Variables:
      CHARACTER(16) :: PNAME = 'INIT_BNDY_COND'
      CHARACTER(96) :: XMSG = ' '

      CHARACTER(16) :: BLNK = ' '

      INTEGER :: S, A2C         ! species index

      CHARACTER(16) :: VNAME    ! variable name

      INTEGER :: JDATE          ! current model date, coded YYYYDDD
      INTEGER :: JTIME          ! current model time, coded HHMMSS

      INTEGER :: DAY_NUM        ! day number

      INTEGER :: IDATE          ! current input file date, coded YYYYDDD

      INTEGER :: STEP_NUM       ! current output time step number

      INTEGER :: STRT, FINI     ! loop limits

      REAL :: BBUF(1:GL_NBNDY, 1:NLAYS)

      CHARACTER(512) :: EQNAME  ! value of an environment variable

      INTERFACE
         SUBROUTINE SCATTER_BNDY_3D(G_ARRAY, L_ARRAY, JDATE, JTIME)
            IMPLICIT NONE
            REAL, INTENT(IN) :: G_ARRAY(:, :, :)
            REAL, INTENT(OUT) :: L_ARRAY(:, :, :)
            INTEGER, INTENT(IN) :: JDATE, JTIME
         END SUBROUTINE SCATTER_BNDY_3D
      END INTERFACE

C-----------------------------------

      IF ( MYPE == 0 ) THEN

C Set input environment variables to point to files for the current day
         DAY_NUM = 1
         CALL SET_INPUT_ENV(SDATE, STIME, DAY_NUM)
         IDATE = SDATE

C Combine the boundary condition files for the different species classes,
C concatenate the combined files in time if the run spans more than one day
C and initialize the boundary condition control variables
         JDATE = SDATE
         JTIME = STIME
         DO STEP_NUM = 0, NSTEPS
            STRT = 1
            FINI = N_GC_ADV
            DO S = STRT, FINI
               VNAME = BCNAME_ADV(S)
               IF ( VNAME .NE. BLNK ) THEN
                  IF ( .NOT. INTERP3(BNDY_GASC_1, VNAME, PNAME,
     &                 JDATE, JTIME, GL_NBNDY*NLAYS, BBUF) ) THEN
                     XMSG = 'Could not read ' // TRIM(VNAME) //
     &                    ' from ' // BNDY_GASC_1
                     CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
                  END IF

                  IF ( .NOT. BWRITE3(ADJ_BC_TS, VNAME, JDATE, JTIME, BBUF) ) THEN
                     XMSG = 'Could not write ' // TRIM(VNAME) //
     &                    ' to ' // TRIM(ADJ_BC_TS)
                     CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
                  END IF

                  A2C = ADV_TO_CNTL_MAP(S)
                  IF ( A2C > 0 ) XS(:, :, A2C, STEP_NUM) = BBUF(:, :)
               END IF
            END DO

            STRT = FINI + 1
            FINI = FINI + N_AE_ADV
            DO S = STRT, FINI
               VNAME = BCNAME_ADV(S)
               IF ( VNAME .NE. BLNK ) THEN
                  IF ( .NOT. INTERP3(BNDY_AERO_1, VNAME, PNAME,
     &                 JDATE, JTIME, GL_NBNDY*NLAYS, BBUF) ) THEN
                     XMSG = 'Could not read ' // TRIM(VNAME) //
     &                    ' from ' // BNDY_AERO_1
                     CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
                  END IF

                  IF ( .NOT. BWRITE3(ADJ_BC_TS, VNAME, JDATE, JTIME, BBUF) ) THEN
                     XMSG = 'Could not write ' // TRIM(VNAME) //
     &                    ' to ' // TRIM(ADJ_BC_TS)
                     CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
                  END IF

                  A2C = ADV_TO_CNTL_MAP(S)
                  IF ( A2C > 0 ) XS(:, :, A2C, STEP_NUM) = BBUF(:, :)
               END IF
            END DO

            STRT = FINI + 1
            FINI = FINI + N_NR_ADV
            DO S = STRT, FINI
               VNAME = BCNAME_ADV(S)
               IF ( VNAME .NE. BLNK ) THEN
                  IF ( .NOT. INTERP3(BNDY_NONR_1, VNAME, PNAME,
     &                 JDATE, JTIME, GL_NBNDY*NLAYS, BBUF) ) THEN
                     XMSG = 'Could not read ' // TRIM(VNAME) //
     &                    ' from ' // BNDY_NONR_1
                     CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
                  END IF

                  IF ( .NOT. BWRITE3(ADJ_BC_TS, VNAME, JDATE, JTIME, BBUF) ) THEN
                     XMSG = 'Could not write ' // TRIM(VNAME) //
     &                    ' to ' // TRIM(ADJ_BC_TS)
                     CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
                  END IF

                  A2C = ADV_TO_CNTL_MAP(S)
                  IF ( A2C > 0 ) XS(:, :, A2C, STEP_NUM) = BBUF(:, :)
               END IF
            END DO

            STRT = FINI + 1
            FINI = FINI + N_TR_ADV
            DO S = STRT, FINI
               VNAME = BCNAME_ADV(S)
               IF ( VNAME .NE. BLNK ) THEN
                  IF ( .NOT. INTERP3(BNDY_TRAC_1, VNAME, PNAME,
     &                 JDATE, JTIME, GL_NBNDY*NLAYS, BBUF) ) THEN
                     XMSG = 'Could not read ' // TRIM(VNAME) //
     &                    ' from ' // BNDY_TRAC_1
                     CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
                  END IF

                  IF ( .NOT. BWRITE3(ADJ_BC_TS, VNAME, JDATE, JTIME, BBUF) ) THEN
                     XMSG = 'Could not write ' // TRIM(VNAME) //
     &                    ' to ' // TRIM(ADJ_BC_TS)
                     CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
                  END IF

                  A2C = ADV_TO_CNTL_MAP(S)
                  IF ( A2C > 0 ) XS(:, :, A2C, STEP_NUM) = BBUF(:, :)
               END IF
            END DO

            CALL NEXTIME(JDATE, JTIME, TSTEP(1))
            IF ( JDATE > IDATE ) THEN
               DAY_NUM = DAY_NUM + 1
               CALL SET_INPUT_ENV(JDATE, JTIME, DAY_NUM)
               IDATE = JDATE
            END IF
         END DO

      END IF                    ! MYPE == 0

C Close the boundary condition files
      IF ( NAME2FID(BNDY_GASC_1) > 0 ) THEN
         IF ( .NOT. CLOSE3(BNDY_GASC_1) ) THEN
            XMSG = 'Could not close ' // TRIM(BNDY_GASC_1)
            CALL M3EXIT(PNAME, SDATE, STIME, XMSG, XSTAT1)
         END IF
      END IF

      IF ( NAME2FID(BNDY_AERO_1) > 0 ) THEN
         IF ( .NOT. CLOSE3(BNDY_AERO_1) ) THEN
            XMSG = 'Could not close ' // TRIM(BNDY_AERO_1)
            CALL M3EXIT(PNAME, SDATE, STIME, XMSG, XSTAT1)
         END IF
      END IF

      IF ( NAME2FID(BNDY_NONR_1) > 0 ) THEN
         IF ( .NOT. CLOSE3(BNDY_NONR_1) ) THEN
            XMSG = 'Could not close ' // TRIM(BNDY_NONR_1)
            CALL M3EXIT(PNAME, SDATE, STIME, XMSG, XSTAT1)
         END IF
      END IF

      IF ( NAME2FID(BNDY_TRAC_1) > 0 ) THEN
         IF ( .NOT. CLOSE3(BNDY_TRAC_1) ) THEN
            XMSG = 'Could not close ' // TRIM(BNDY_TRAC_1)
            CALL M3EXIT(PNAME, SDATE, STIME, XMSG, XSTAT1)
         END IF
      END IF

C Save the initial values of the boundary condition control variables as
C "background" values
      JDATE = SDATE
      JTIME = STIME
      DO STEP_NUM = 0, NSTEPS
         CALL SCATTER_BNDY_3D(
     &        XS(:, :, :, STEP_NUM),
     &        BG_BC_TS(:, :, :, STEP_NUM),
     &        JDATE, JTIME)
         CALL NEXTIME(JDATE, JTIME, TSTEP(1))
      END DO

C Reset the paths to the boundary condition files to the path to ADJ_BC_TS
      CALL NAMEVAL(ADJ_BC_TS, EQNAME)
      CALL SETENVVAR(BNDY_GASC_1, EQNAME)
      CALL SETENVVAR(BNDY_AERO_1, EQNAME)
      CALL SETENVVAR(BNDY_NONR_1, EQNAME)
      CALL SETENVVAR(BNDY_TRAC_1, EQNAME)

      END SUBROUTINE INIT_BNDY_COND
