
C***********************************************************************
C   Portions of Models-3/CMAQ software were developed or based on      *
C   information from various groups: Federal Government employees,     *
C   contractors working on a United States Government contract, and    *
C   non-Federal sources (including research institutions).  These      *
C   research institutions have given the Government permission to      *
C   use, prepare derivative works, and distribute copies of their      *
C   work in Models-3/CMAQ to the public and to permit others to do     *
C   so.  EPA therefore grants similar permissions for use of the       *
C   Models-3/CMAQ software, but users are requested to provide copies  *
C   of derivative works to the Government without restrictions as to   *
C   use by others.  Users are responsible for acquiring their own      *
C   copies of commercial software associated with Models-3/CMAQ and    *
C   for complying with vendor requirements.  Software copyrights by    *
C   the MCNC Environmental Modeling Center are used with their         *
C   permissions subject to the above restrictions.                     *
C***********************************************************************

C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      SUBROUTINE GET_BG_FG(LGRID, JDATE, JTIME, F)

C-----------------------------------------------------------------------
C Function:
C   Get the function value and gradient of the background operator at the
C   initial time

C Revision History:
C   Apr 2007 Kumaresh Singh and Adrian Sandu at Virginia Tech: created
C   Jun 2009 Peter Percell at UH-IMAQS: rewrote
C-----------------------------------------------------------------------

      USE GRID_CONF             ! horizontal & vertical domain specifications
      USE FOURDVAR_MOD          ! 4DVAR data
      USE ADJOINT_FILES         ! adjoint file data and functions

      IMPLICIT NONE

C Include Files:
#ifdef parallel
      INCLUDE SUBST_MPICH       ! MPI definitions and parameters
#endif

C Parameters:
      REAL(8), PARAMETER :: BG_WGT = 12.0D0

C Arguments:
      REAL, POINTER :: LGRID(:, :, :, :) ! gradient of cost function
      INTEGER, INTENT(IN) :: JDATE ! current model date, coded YYYYDDD
      INTEGER, INTENT(IN) :: JTIME ! current model time, coded HHMMSS
      REAL(8), INTENT(INOUT) :: F ! cumulative value of cost function (for local domain)

C Local Variables:
      CHARACTER(16) :: PNAME = 'GET_BG_FG'
      CHARACTER(96) :: XMSG = ' '

      REAL(8) :: STD(1:NLAYS, 1:N_CNTL_SPC) ! (global) standard deviation

      INTEGER :: C, R, L, S, CS ! column, row, layer and species indices

      REAL(8) :: D

#ifdef parallel
      REAL(8) :: STD_LD         ! (local) standard deviation
      INTEGER :: IERROR         ! MPI error code
#endif

C-----------------------------------------------------------------------

C Estimate the (global) standard deviation of each background value over each
C layer
      DO S = 1, N_CNTL_SPC
         DO L = 1, NLAYS
#ifdef parallel
            STD_LD = SUM(DBLE(BG_GRID(1:NCOLS, 1:NROWS, L, S)))
            CALL MPI_ALLREDUCE(STD_LD, STD(L, S), 1, MPI_REAL8, MPI_SUM,
     &           MPI_COMM_WORLD, IERROR)
#else
            STD(L, S) = SUM(DBLE(BG_GRID(1:NCOLS, 1:NROWS, L, S)))
#endif
            STD(L, S) = BG_WGT * STD(L, S) / (GL_NCOLS * GL_NROWS)
            IF ( ABS(STD(L, S)) < 1.0D-30 ) STD(L, S) = 1.0D0
         END DO
      END DO

C Update the cost function F and gradient LGRID
      DO S = 1, N_CNTL_SPC
         CS = CNTL_TO_CGRID_MAP(S)
         DO L = 1, NLAYS
            DO R = 1, NROWS
               DO C = 1, NCOLS
                  D = (CNTL_GRID(C, R, L, S) - BG_GRID(C, R, L, S)) / STD(L, S)
                  F = F + D**2 / 2.0D0
                  LGRID(C, R, L, CS) = LGRID(C, R, L, CS) + D / STD(L, S)
               END DO
            END DO
         END DO
      END DO

      END SUBROUTINE GET_BG_FG
