C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

      MODULE FOURDVAR_MOD

C-----------------------------------------------------------------------
C Function:
C   Define and initialize FOURDVAR data

C Revision History:
C   Aug 2009 by Peter Percell at UH-IMAQS: Created
C-----------------------------------------------------------------------

      IMPLICIT NONE
      SAVE

C-----------------------------------
C Data
C-----------------------------------

C Number of observed species
      INTEGER :: N_OBS_SPC

C Observed species names and units
      CHARACTER(16), ALLOCATABLE :: OBS_SPC_NAME(:)
      CHARACTER(16), ALLOCATABLE :: OBS_SPC_UNITS(:)

C Map from the observed species list to the CGRID species list
      INTEGER, ALLOCATABLE :: OBS_TO_CGRID_MAP(:)

C Number of control species
      INTEGER :: N_CNTL_SPC

C Control species names and units
      CHARACTER(16), ALLOCATABLE :: CNTL_SPC_NAME(:)
      CHARACTER(16), ALLOCATABLE :: CNTL_SPC_UNITS(:)

C Maps between the control species list and the CGRID, advected and diffused
C species lists
      INTEGER, ALLOCATABLE :: CNTL_TO_CGRID_MAP(:)
      INTEGER, ALLOCATABLE :: CNTL_TO_ADV_MAP(:)
      INTEGER, ALLOCATABLE :: ADV_TO_CNTL_MAP(:)
      INTEGER, ALLOCATABLE :: CNTL_TO_DIFF_MAP(:)
      INTEGER, ALLOCATABLE :: DIFF_TO_CNTL_MAP(:)

C Background values of control variables
      REAL, ALLOCATABLE :: BG_GRID(:, :, :, :)

C Current values of control variables
      REAL, ALLOCATABLE :: CNTL_GRID(:, :, :, :)

C Emissions scaling factor
      REAL, ALLOCATABLE :: EM_SF(:, :, :, :)

C Gradient of cost function wrt emissions scaling factor
      REAL, ALLOCATABLE :: LGRID_EM_SF(:, :, :, :)

C Boundary condition scaling factor
      REAL, ALLOCATABLE :: BC_SF(:, :, :)

C Gradient of cost function wrt boundary condition scaling factor
      REAL, ALLOCATABLE :: LGRID_BC_SF(:, :, :)

C Background values of control variables in boundary condition time series
      REAL, ALLOCATABLE :: BG_BC_TS(:, :, :, :)

C Current values of control variables in boundary condition time series
      REAL, ALLOCATABLE :: CNTL_BC_TS(:, :, :, :)

C Gradient of cost function wrt control variables in boundary condition time
C series
      REAL, ALLOCATABLE :: LGRID_BC_TS(:, :, :, :)

C Map from advected species to BC species names
      CHARACTER(16), ALLOCATABLE :: BCNAME_ADV(:)

C Map from control species to BC species names
      CHARACTER(16), ALLOCATABLE :: BCNAME_CNTL(:)

C$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

      CONTAINS

C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C Initialize data
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      SUBROUTINE FOURDVAR_INIT(JDATE, JTIME)

      USE CGRID_SPCS            ! CGRID species number and offsets

      IMPLICIT NONE

C Include Files:
      INCLUDE SUBST_GC_SPC      ! gas chemistry species table
      INCLUDE SUBST_AE_SPC      ! aerosol species table
      INCLUDE SUBST_NR_SPC      ! non-reactive species table
      INCLUDE SUBST_TR_SPC      ! tracer species table
      INCLUDE SUBST_GC_ADV      ! gas chem advection species and map table
      INCLUDE SUBST_AE_ADV      ! aerosol advection species and map table
      INCLUDE SUBST_NR_ADV      ! non-react advection species and map table
      INCLUDE SUBST_TR_ADV      ! tracer advection species and map table
      INCLUDE SUBST_GC_DIFF     ! gas chem diffusion species and map table
      INCLUDE SUBST_AE_DIFF     ! aerosol diffusion species and map table
      INCLUDE SUBST_NR_DIFF     ! non-react diffusion species and map table
      INCLUDE SUBST_TR_DIFF     ! tracer diffusion species and map table
      INCLUDE SUBST_IOPARMS     ! I/O parameters definitions

C Arguments:
      INTEGER, INTENT(IN) :: JDATE ! starting date (YYYYDDD)
      INTEGER, INTENT(IN) :: JTIME ! starting time (HHMMSS)

C External Functions (not already declared by IODECL3.EXT):
      INTEGER, EXTERNAL :: INDEX1

C Local Variables:
      CHARACTER(16) :: PNAME = 'FOURDVAR_INIT'
      CHARACTER(96) :: XMSG = ' '

      CHARACTER(16) :: SPC_LIST(N_GC_SPC + N_AE_SPC + N_NR_SPC + N_TR_SPC)
      LOGICAL, ALLOCATABLE :: SPC_KNOWN(:)
      INTEGER :: N_UNKNOWN

      INTEGER :: ALLOCSTAT

      INTEGER :: S              ! species index
      INTEGER :: STRT           ! loop limit
      INTEGER :: INDX

      INTERFACE
         SUBROUTINE GET_ENVLIST(ENV_VAR, NVARS, V_LIST)
            IMPLICIT NONE
            CHARACTER(*), INTENT(IN) :: ENV_VAR
            INTEGER, INTENT(OUT) :: NVARS
            CHARACTER(16), INTENT(OUT) :: V_LIST(:)
         END SUBROUTINE GET_ENVLIST
      END INTERFACE

C-----------------------------------------------------------------------

C Get CGRID offsets
      CALL CGRID_MAP(NSPCSD, GC_STRT, AE_STRT, NR_STRT, TR_STRT)

C Get the list of observed species
      CALL GET_ENVLIST('OBS_SPCS', N_OBS_SPC, SPC_LIST)

      IF ( N_OBS_SPC == 1 .AND. TRIM(SPC_LIST(1)) == 'ALL' ) THEN
         N_OBS_SPC = N_GC_SPC + N_AE_SPC + N_NR_SPC + N_TR_SPC ! everything

         STRT = 0
         DO INDX = 1, N_GC_SPC
            SPC_LIST(STRT + INDX) = GC_SPC(INDX)
         END DO

         STRT = N_GC_SPC
         DO INDX = 1, N_AE_SPC
            SPC_LIST(STRT + INDX) = AE_SPC(INDX)
         END DO

         STRT = N_GC_SPC + N_AE_SPC
         DO INDX = 1, N_NR_SPC
            SPC_LIST(STRT + INDX) = NR_SPC(INDX)
         END DO

         STRT = N_GC_SPC + N_AE_SPC + N_NR_SPC
         DO INDX = 1, N_TR_SPC
            SPC_LIST(STRT + INDX) = TR_SPC(INDX)
         END DO
      END IF

      IF ( N_OBS_SPC > 0 ) THEN

C Allocate memory for the map from the observed species list to the CGRID
C species list and for the lists of observed species names and units
         ALLOCATE(
     &        OBS_TO_CGRID_MAP(1:N_OBS_SPC),
     &        OBS_SPC_NAME(1:N_OBS_SPC),
     &        OBS_SPC_UNITS(1:N_OBS_SPC),
     &        STAT = ALLOCSTAT)
         IF ( ALLOCSTAT /= 0 ) THEN
            XMSG = 'Failure allocating memory for observed species data'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT2)
         END IF
         OBS_TO_CGRID_MAP(:) = 0

C Create map from the observed species list to the CGRID species list and
C create the lists of observed species names and units
         ALLOCATE(
     &        SPC_KNOWN(1:N_OBS_SPC),
     &        STAT = ALLOCSTAT)
         IF ( ALLOCSTAT /= 0 ) THEN
            XMSG = 'Failure allocating memory for SPC_KNOWN'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT2)
         END IF
         SPC_KNOWN = .FALSE.

         DO S = 1, N_OBS_SPC
            INDX = INDEX1(SPC_LIST(S), N_GC_SPC, GC_SPC)
            IF ( INDX > 0 ) THEN
               OBS_TO_CGRID_MAP(S) = GC_STRT + INDX - 1
               OBS_SPC_NAME(S) = SPC_LIST(S)
               OBS_SPC_UNITS(S) = 'ppmV'
               SPC_KNOWN(S) = .TRUE.
            END IF
         END DO

         DO S = 1, N_OBS_SPC
            INDX = INDEX1(SPC_LIST(S), N_AE_SPC, AE_SPC)
            IF ( INDX > 0 ) THEN
               OBS_TO_CGRID_MAP(S) = AE_STRT + INDX - 1
               OBS_SPC_NAME(S) = SPC_LIST(S)
               IF ( OBS_SPC_NAME(S)(1:3) == 'NUM' ) THEN
                  OBS_SPC_UNITS(S) = 'number/m**3'
               ELSE IF ( OBS_SPC_NAME(S)(1:3) == 'SRF' ) THEN
                  OBS_SPC_UNITS(S) = 'm**2/m**3'
               ELSE
                  OBS_SPC_UNITS(S) = 'microg/m**3'
               END IF
               SPC_KNOWN(S) = .TRUE.
            END IF
         END DO

         DO S = 1, N_OBS_SPC
            INDX = INDEX1(SPC_LIST(S), N_NR_SPC, NR_SPC)
            IF ( INDX > 0 ) THEN
               OBS_TO_CGRID_MAP(S) = NR_STRT + INDX - 1
               OBS_SPC_NAME(S) = SPC_LIST(S)
               OBS_SPC_UNITS(S) = 'ppmV'
               SPC_KNOWN(S) = .TRUE.
            END IF
         END DO

         DO S = 1, N_OBS_SPC
            INDX = INDEX1(SPC_LIST(S), N_TR_SPC, TR_SPC)
            IF ( INDX > 0 ) THEN
               OBS_TO_CGRID_MAP(S) = TR_STRT + INDX - 1
               OBS_SPC_NAME(S) = SPC_LIST(S)
               OBS_SPC_UNITS(S) = 'ppmV'
               SPC_KNOWN(S) = .TRUE.
            END IF
         END DO

         N_UNKNOWN = 0
         DO S = 1, N_OBS_SPC
            IF ( .NOT. SPC_KNOWN(S) ) THEN
               XMSG = 'Species ' // TRIM(SPC_LIST(S))
     &              // ' is unknown for this model'
               CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
               N_UNKNOWN = N_UNKNOWN + 1
            END IF
         END DO
         IF ( N_UNKNOWN > 0 ) CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT3)

      END IF                    ! N_OBS_SPC > 0

C Get the list of control species
      CALL GET_ENVLIST('CNTL_SPCS', N_CNTL_SPC, SPC_LIST)

      IF ( N_CNTL_SPC == 1 .AND. TRIM(SPC_LIST(1)) == 'ALL' ) THEN
         N_CNTL_SPC = N_GC_SPC + N_AE_SPC + N_NR_SPC + N_TR_SPC ! everything

         STRT = 0
         DO INDX = 1, N_GC_SPC
            SPC_LIST(STRT + INDX) = GC_SPC(INDX)
         END DO

         STRT = N_GC_SPC
         DO INDX = 1, N_AE_SPC
            SPC_LIST(STRT + INDX) = AE_SPC(INDX)
         END DO

         STRT = N_GC_SPC + N_AE_SPC
         DO INDX = 1, N_NR_SPC
            SPC_LIST(STRT + INDX) = NR_SPC(INDX)
         END DO

         STRT = N_GC_SPC + N_AE_SPC + N_NR_SPC
         DO INDX = 1, N_TR_SPC
            SPC_LIST(STRT + INDX) = TR_SPC(INDX)
         END DO
      END IF

      IF ( N_CNTL_SPC > 0 ) THEN

C Allocate memory for maps between the control species list and the CGRID,
C advected and diffused species lists and for the lists of control species
C names and units
         ALLOCATE(
     &        CNTL_TO_CGRID_MAP(1:N_CNTL_SPC),
     &        CNTL_TO_ADV_MAP(1:N_CNTL_SPC),
     &        ADV_TO_CNTL_MAP(N_GC_ADV+N_AE_ADV+N_NR_ADV+N_TR_ADV+1),
     &        CNTL_TO_DIFF_MAP(1:N_CNTL_SPC),
     &        DIFF_TO_CNTL_MAP(N_GC_DIFF+N_AE_DIFF+N_NR_DIFF+N_TR_DIFF),
     &        CNTL_SPC_NAME(1:N_CNTL_SPC),
     &        CNTL_SPC_UNITS(1:N_CNTL_SPC),
     &        STAT = ALLOCSTAT)
         IF ( ALLOCSTAT /= 0 ) THEN
            XMSG = 'Failure allocating memory for control species data'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT2)
         END IF
         CNTL_TO_CGRID_MAP(:) = 0
         CNTL_TO_ADV_MAP(:) = 0
         ADV_TO_CNTL_MAP(:) = 0
         CNTL_TO_DIFF_MAP(:) = 0
         DIFF_TO_CNTL_MAP(:) = 0

C Create map from the control species list to the CGRID species list and
C create the lists of control species names and units
         ALLOCATE(
     &        SPC_KNOWN(1:N_CNTL_SPC),
     &        STAT = ALLOCSTAT)
         IF ( ALLOCSTAT /= 0 ) THEN
            XMSG = 'Failure allocating memory for SPC_KNOWN'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT2)
         END IF
         SPC_KNOWN = .FALSE.

         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(SPC_LIST(S), N_GC_SPC, GC_SPC)
            IF ( INDX > 0 ) THEN
               CNTL_TO_CGRID_MAP(S) = GC_STRT + INDX - 1
               CNTL_SPC_NAME(S) = SPC_LIST(S)
               CNTL_SPC_UNITS(S) = 'ppmV'
               SPC_KNOWN(S) = .TRUE.
            END IF
         END DO

         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(SPC_LIST(S), N_AE_SPC, AE_SPC)
            IF ( INDX > 0 ) THEN
               CNTL_TO_CGRID_MAP(S) = AE_STRT + INDX - 1
               CNTL_SPC_NAME(S) = SPC_LIST(S)
               IF ( CNTL_SPC_NAME(S)(1:3) == 'NUM' ) THEN
                  CNTL_SPC_UNITS(S) = 'number/m**3'
               ELSE IF ( CNTL_SPC_NAME(S)(1:3) == 'SRF' ) THEN
                  CNTL_SPC_UNITS(S) = 'm**2/m**3'
               ELSE
                  CNTL_SPC_UNITS(S) = 'microg/m**3'
               END IF
               SPC_KNOWN(S) = .TRUE.
            END IF
         END DO

         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(SPC_LIST(S), N_NR_SPC, NR_SPC)
            IF ( INDX > 0 ) THEN
               CNTL_TO_CGRID_MAP(S) = NR_STRT + INDX - 1
               CNTL_SPC_NAME(S) = SPC_LIST(S)
               CNTL_SPC_UNITS(S) = 'ppmV'
               SPC_KNOWN(S) = .TRUE.
            END IF
         END DO

         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(SPC_LIST(S), N_TR_SPC, TR_SPC)
            IF ( INDX > 0 ) THEN
               CNTL_TO_CGRID_MAP(S) = TR_STRT + INDX - 1
               CNTL_SPC_NAME(S) = SPC_LIST(S)
               CNTL_SPC_UNITS(S) = 'ppmV'
               SPC_KNOWN(S) = .TRUE.
            END IF
         END DO

         N_UNKNOWN = 0
         DO S = 1, N_CNTL_SPC
            IF ( .NOT. SPC_KNOWN(S) ) THEN
               XMSG = 'Species ' // TRIM(SPC_LIST(S))
     &              // ' is unknown for this model'
               CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
               N_UNKNOWN = N_UNKNOWN + 1
            END IF
         END DO
         IF ( N_UNKNOWN > 0 ) CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT3)

C Create maps between the control species list and the advected species list
         STRT = 0
         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(CNTL_SPC_NAME(S), N_GC_ADV, GC_ADV)
            IF ( INDX > 0 ) THEN
               CNTL_TO_ADV_MAP(S) = STRT + INDX
               ADV_TO_CNTL_MAP(STRT + INDX) = S
            END IF
         END DO

         STRT = N_GC_ADV
         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(CNTL_SPC_NAME(S), N_AE_ADV, AE_ADV)
            IF ( INDX > 0 ) THEN
               CNTL_TO_ADV_MAP(S) = STRT + INDX
               ADV_TO_CNTL_MAP(STRT + INDX) = S
            END IF
         END DO

         STRT = N_GC_ADV + N_AE_ADV
         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(CNTL_SPC_NAME(S), N_NR_ADV, NR_ADV)
            IF ( INDX > 0 ) THEN
               CNTL_TO_ADV_MAP(S) = STRT + INDX
               ADV_TO_CNTL_MAP(STRT + INDX) = S
            END IF
         END DO

         STRT = N_GC_ADV + N_AE_ADV + N_NR_ADV
         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(CNTL_SPC_NAME(S), N_TR_ADV, TR_ADV)
            IF ( INDX > 0 ) THEN
               CNTL_TO_ADV_MAP(S) = STRT + INDX
               ADV_TO_CNTL_MAP(STRT + INDX) = S
            END IF
         END DO

C Create maps between the control species list and the diffused species list
         STRT = 0
         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(CNTL_SPC_NAME(S), N_GC_DIFF, GC_DIFF)
            IF ( INDX > 0 ) THEN
               CNTL_TO_DIFF_MAP(S) = STRT + INDX
               DIFF_TO_CNTL_MAP(STRT + INDX) = S
            END IF
         END DO

         STRT = N_GC_DIFF
         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(CNTL_SPC_NAME(S), N_AE_DIFF, AE_DIFF)
            IF ( INDX > 0 ) THEN
               CNTL_TO_DIFF_MAP(S) = STRT + INDX
               DIFF_TO_CNTL_MAP(STRT + INDX) = S
            END IF
         END DO

         STRT = N_GC_DIFF + N_AE_DIFF
         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(CNTL_SPC_NAME(S), N_NR_DIFF, NR_DIFF)
            IF ( INDX > 0 ) THEN
               CNTL_TO_DIFF_MAP(S) = STRT + INDX
               DIFF_TO_CNTL_MAP(STRT + INDX) = S
            END IF
         END DO

         STRT = N_GC_DIFF + N_AE_DIFF + N_NR_DIFF
         DO S = 1, N_CNTL_SPC
            INDX = INDEX1(CNTL_SPC_NAME(S), N_TR_DIFF, TR_DIFF)
            IF ( INDX > 0 ) THEN
               CNTL_TO_DIFF_MAP(S) = STRT + INDX
               DIFF_TO_CNTL_MAP(STRT + INDX) = S
            END IF
         END DO

      END IF                    ! N_CNTL_SPC > 0

      END SUBROUTINE FOURDVAR_INIT

C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      END MODULE FOURDVAR_MOD
