
C***********************************************************************
C   Portions of Models-3/CMAQ software were developed or based on      *
C   information from various groups: Federal Government employees,     *
C   contractors working on a United States Government contract, and    *
C   non-Federal sources (including research institutions).  These      *
C   research institutions have given the Government permission to      *
C   use, prepare derivative works, and distribute copies of their      *
C   work in Models-3/CMAQ to the public and to permit others to do     *
C   so.  EPA therefore grants similar permissions for use of the       *
C   Models-3/CMAQ software, but users are requested to provide copies  *
C   of derivative works to the Government without restrictions as to   *
C   use by others.  Users are responsible for acquiring their own      *
C   copies of commercial software associated with Models-3/CMAQ and    *
C   for complying with vendor requirements.  Software copyrights by    *
C   the MCNC Environmental Modeling Center are used with their         *
C   permissions subject to the above restrictions.                     *
C***********************************************************************

C RCS file, release, date & time of last delta, author, state, [and locker]
C $Header: /home/aqm/ppercell/aq/cmaq_4.6/models/CCTM/src/driver/ctm/advstep.F,v 1.1.1.1 2006/09/22 18:19:30 sjr Exp $

C what(1) key, module and SID; SCCS file; date and time of last delta:
C %W% %P% %G% %U%

C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      SUBROUTINE ADVSTEP(JDATE, JTIME, TSTEP, NADVSS)

C-----------------------------------------------------------------------
C Function:
C    Use Courant-number conditions to determine the number of advection
C    sub-steps of the synchronization step needed in each layer.

C Preconditions:
C    The synchronization step, TSTEP(2), has already been determined.

C Revision history:
C   01/20/09 by Peter Percell at UH-IMAQS:
C   - Rewrote for use with adjoint/4dvar code
C-----------------------------------------------------------------------

      USE GRID_CONF             ! horizontal & vertical domain specifications
      USE SUBST_MODULES         ! stenex

      IMPLICIT NONE

C Includes:
      INCLUDE SUBST_CONST       ! constants
      INCLUDE SUBST_IOPARMS     ! I/O parameters definitions
      INCLUDE SUBST_IOFDESC     ! file header data structure
      INCLUDE SUBST_IODECL      ! I/O definitions and declarations
      INCLUDE SUBST_FILES_ID    ! file name parameters

C Parameters:
c$$$      REAL, PARAMETER :: CC = 0.75 ! maximum Courant number allowed
c$$$      REAL, PARAMETER :: EPS = 0.2
      REAL, PARAMETER :: CC = 0.90 ! maximum Courant number allowed
      REAL, PARAMETER :: EPS = 0.05

C Arguments:
      INTEGER, INTENT(IN) :: JDATE ! current model simulation date (YYYYDDD)
      INTEGER, INTENT(IN) :: JTIME ! current model simulation time (HHMMSS)
      INTEGER, INTENT(IN) :: TSTEP(2) ! time step vector (HHMMSS)
                                !   TSTEP(1) = local output step
                                !   TSTEP(2) = sciproc sync. step (chem)
      INTEGER, INTENT(OUT) :: NADVSS(:) ! number of advection sub-steps per sync step

C External Functions (not already declared by IODECL3.EXT):
      INTEGER, EXTERNAL :: TIME2SEC, SECSDIFF
      LOGICAL, EXTERNAL :: CURRSTEP

C Local Variables:
      LOGICAL, SAVE :: FIRSTIME = .TRUE.
      INTEGER, SAVE :: LOGDEV   ! unit number for log device

      INTEGER, SAVE :: WSTEP = 0 ! wind file interpolation time step
      INTEGER, SAVE :: WSDATE, WSTIME ! wind file start date and time
      INTEGER, SAVE :: FDATE, FTIME ! wind file current date and time
      INTEGER, SAVE :: FSTEP    ! wind file time step

      REAL, SAVE :: IDX1        ! 1/dx1
      REAL, SAVE :: IDX2        ! 1/dx2

      REAL :: WIND(1:NCOLS+1, 1:NROWS+1) ! generic contravariant wind velocity

      INTEGER :: C, R, K, L     ! loop counters
      INTEGER :: EDATE, ETIME
      INTEGER :: SYNC

      REAL :: MXADV

      REAL :: WIND_DSI(1:NLAYS) ! max { component velocity / dXi }

      CHARACTER(16) :: PNAME = 'ADVSTEP'
      CHARACTER(16) :: UORV
      CHARACTER(96) :: XMSG = ' '

      INTEGER :: MY_TEMP
      INTEGER, SAVE :: STARTCOL, ENDCOL
      INTEGER, SAVE :: STARTROW, ENDROW

      INTERFACE
         SUBROUTINE HCONTVEL(FDATE, FTIME, FSTEP, LVL, UORV, WIND)
            IMPLICIT NONE
            INTEGER, INTENT(IN)       :: FDATE, FTIME, FSTEP
            INTEGER, INTENT(IN OUT)   :: LVL
            CHARACTER(16), INTENT(IN) :: UORV
            REAL, INTENT(OUT)         :: WIND(:, :)
         END SUBROUTINE HCONTVEL
      END INTERFACE

C-----------------------------------------------------------------------

      IF ( FIRSTIME ) THEN

         FIRSTIME = .FALSE.

         LOGDEV = INIT3()

C Open wind field file and get header data
         IF ( .NOT. OPEN3(MET_DOT_3D, FSREAD3, PNAME) ) THEN
            XMSG = 'Could not open ' // MET_DOT_3D // ' file'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
         END IF

         IF ( .NOT. DESC3(MET_DOT_3D) ) THEN
            XMSG = 'Could not get ' // MET_DOT_3D // ' file description'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
         END IF

C Get the wind file start date and time and the wind file time step
         WSDATE = SDATE3D
         WSTIME = STIME3D
         FSTEP = TSTEP3D

C Get cell size in meters
         IF ( GDTYP_GD .EQ. LATGRD3 ) THEN
            IDX2 = 1.0 / (DG2M * YCELL_GD)
            IDX1 = 1.0 / (DG2M * XCELL_GD
     &           * COS(PI180 * (YORIG_GD + YCELL_GD * FLOAT(NROWS / 2))))
         ELSE
            IDX1 = 1.0 / XCELL_GD
            IDX2 = 1.0 / YCELL_GD
         END IF

         CALL SUBST_LOOP_INDEX('C', 1, NCOLS, 1, MY_TEMP,
     &        STARTCOL, ENDCOL)

         CALL SUBST_LOOP_INDEX('R', 1, NROWS, 1, MY_TEMP,
     &        STARTROW, ENDROW)

      END IF                    ! FIRSTIME

      WRITE(LOGDEV, 92013) JDATE, JTIME

C Get the wind file start date and time for this output step; do not interpolate
      IF ( .NOT. CURRSTEP(JDATE, JTIME, WSDATE, WSTIME, FSTEP,
     &     FDATE, FTIME) ) THEN
         XMSG = 'Cannot get starting date and time'
         CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT3)
      END IF

C Establish end time for Courant number calculation
      EDATE = JDATE
      ETIME = JTIME
      CALL NEXTIME(EDATE, ETIME, TSTEP(1))

      WIND_DSI(:) = -1.0
      DO WHILE ( SECSDIFF(FDATE, FTIME, EDATE, ETIME) .GE. 0 )
         DO L = 1, NLAYS
            UORV = 'X1VEL'
            CALL HCONTVEL(FDATE, FTIME, WSTEP, L, UORV, WIND)
            DO R = 1, NROWS
               DO C = STARTCOL, ENDCOL
                  WIND_DSI(L) = MAX(WIND_DSI(L), ABS(WIND(C, R)) * IDX1)
               END DO
            END DO

            UORV = 'X2VEL'
            CALL HCONTVEL(FDATE, FTIME, WSTEP, L, UORV, WIND)
            DO R = STARTROW, ENDROW
               DO C = 1, NCOLS
                  WIND_DSI(L) = MAX(WIND_DSI(L), ABS(WIND(C, R)) * IDX2)
               END DO
            END DO
         END DO

         IF ( SECSDIFF(EDATE, ETIME, FDATE, FTIME) .GE. 0 ) THEN
            EXIT
         ELSE
            CALL NEXTIME(FDATE, FTIME, FSTEP) ! next wind file time
         END IF
      END DO

      DO L = 1, NLAYS
         WIND_DSI(L) = SUBST_GLOBAL_MAX(WIND_DSI(L))
      END DO

C For each layer, determine the smallest number of advection sub-steps required
C by the Courant condition
      WRITE(LOGDEV, 92025)
      SYNC = TIME2SEC(TSTEP(2))
      DO L = NLAYS, 1, -1
         MXADV = CC / WIND_DSI(L)
         IF ( MXADV .LT. SYNC ) THEN
            NADVSS(L) = CEILING(SYNC / MXADV)
         ELSE
            NADVSS(L) = 1
         END IF
         WRITE(LOGDEV, 92027) L, REAL(SYNC) / NADVSS(L), NADVSS(L)

C Note that the associated advection time step,
C		DT = REAL(SYNC) / REAL(NADVSS(L))
C does not need to be an integer; it will be rounded off to either FLOOR(DT) or
C CEILING(DT) when necessary, e.g., to be compatible with IO/API functions.
C
C Definition (Courant Number): CN(WIND_DSI(L), DT) = WIND_DSI(L) * DT.
C
C Thus, if CN(WIND_DSI(L), DT) <= CC < 1 and WIND_DSI(L) <= EPS < 1 - CC, then
C		CN(WIND_DSI(L), CEILING(DT)) < CC + EPS < 1
C For example, if CC = .75 and EPS = 0.2, then
C		CN(WIND_DSI(L), CEILING(DT)) < .95
         IF ( MOD(SYNC, NADVSS(L)) .NE. 0 .AND.
     &        WIND_DSI(L) .GT. EPS ) THEN
            XMSG = 'The effective Courant Number may be too large'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT3)
         END IF
      END DO

      WRITE(LOGDEV, *) ' '

      RETURN

C------------------  FORMAT  STATEMENTS   ------------------------------

92013 FORMAT(/ 5X, 'From ADVSTEP - date/time: ', I8, '/', I6.6)

92025 FORMAT(/ 5X, 'Layer', 3X, 'Advection', 3X, 'per Sync'
     &     /13X, 'Step (SEC)', 3X, 'Step')

92027 FORMAT(5X, I4, 6X, F6.2, 6X, I2)

      END
