C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      SUBROUTINE GATHER_BNDY_3D(L_ARRAY, G_ARRAY, JDATE, JTIME)

C-----------------------------------------------------------------------
C Function:
C   Gather 3D boundary arrays from local processors in a global array on
C   processor 0

C Revision History:
C   Jul 2009 Peter Percell at UH-IMAQS: created
C-----------------------------------------------------------------------

      USE GRID_CONF             ! horizontal & vertical domain specifications

      IMPLICIT NONE

C Includes:
      INCLUDE SUBST_IOPARMS     ! I/O parameters definitions (XSTAT*)
#ifdef parallel
      INCLUDE SUBST_MPICH       ! MPI definitions and parameters
#endif

C Parameters:
      INTEGER, PARAMETER :: TAG1 = 901 ! MPI message tag for processor ID
      INTEGER, PARAMETER :: TAG2 = 902 ! MPI message tag for data array

C Arguments:
      REAL, INTENT(IN) :: L_ARRAY(:, :, :) ! local 3D boundary array
      REAL, INTENT(OUT) :: G_ARRAY(:, :, :) ! global 3D boundary array
      INTEGER, INTENT(IN) :: JDATE ! date, coded YYYYDDD
      INTEGER, INTENT(IN) :: JTIME ! time, coded HHMMSS

C Local Variables:
#ifdef parallel
      LOGICAL, SAVE :: FIRSTIME = .TRUE.

      CHARACTER(16) :: PNAME = 'GATHER_BNDY_3D'
      CHARACTER(96) :: XMSG = ' '

      INTEGER, SAVE :: NPROCS   ! total number of processors

      REAL, ALLOCATABLE :: RECV(:, :, :)
      INTEGER :: ALLOCSTAT

      INTEGER :: K              ! loop counter

      INTEGER :: WHO            ! rank of sending processor
      INTEGER :: ISTATUS(MPI_STATUS_SIZE) ! MPI status code
      INTEGER :: IERROR         ! MPI error code
      LOGICAL :: LERROR         ! local error
      LOGICAL :: RERROR         ! local MPI ALLREDUCE error

      LOGICAL, ALLOCATABLE, SAVE :: S_BND_PE(:), E_BND_PE(:), N_BND_PE(:), W_BND_PE(:)

      INTEGER, ALLOCATABLE, SAVE :: GL_CZ(:), GL_RZ(:) ! column, row offsets in global domain
      INTEGER, ALLOCATABLE, SAVE :: NC(:), NR(:) ! number of columns, rows in local domain
      INTEGER, ALLOCATABLE, SAVE :: NB(:) ! number of boundary cells for local domain

      INTEGER, ALLOCATABLE, SAVE :: SZ(:), GL_SZ(:) ! local, global zero index for south boundary
      INTEGER, ALLOCATABLE, SAVE :: EZ(:), GL_EZ(:) ! local, global zero index for east boundary
      INTEGER, ALLOCATABLE, SAVE :: NZ(:), GL_NZ(:) ! local, global zero index for north boundary
      INTEGER, ALLOCATABLE, SAVE :: WZ(:), GL_WZ(:) ! local, global zero index for west boundary
#endif

C-----------------------------------------------------------------------

#ifdef parallel

      IF ( FIRSTIME ) THEN
         FIRSTIME = .FALSE.

C Get number of processors
         CALL MPI_COMM_SIZE(MPI_COMM_WORLD, NPROCS, IERROR)
         IF ( IERROR /= 0 ) THEN
            XMSG = 'Error in MPI_COMM_SIZE'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT3)
         END IF

         ALLOCATE(
     &        S_BND_PE(0:NPROCS-1),
     &        E_BND_PE(0:NPROCS-1),
     &        N_BND_PE(0:NPROCS-1),
     &        W_BND_PE(0:NPROCS-1),
     &        GL_CZ(0:NPROCS-1),
     &        GL_RZ(0:NPROCS-1),
     &        NC(0:NPROCS-1),
     &        NR(0:NPROCS-1),
     &        NB(0:NPROCS-1),
     &        SZ(0:NPROCS-1),
     &        GL_SZ(0:NPROCS-1),
     &        EZ(0:NPROCS-1),
     &        GL_EZ(0:NPROCS-1),
     &        NZ(0:NPROCS-1),
     &        GL_NZ(0:NPROCS-1),
     &        WZ(0:NPROCS-1),
     &        GL_WZ(0:NPROCS-1),
     &        STAT = ALLOCSTAT)
         IF ( ALLOCSTAT /= 0 ) THEN
            XMSG = 'Failure allocating boundary location arrays'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
         END IF

         DO K = 0, NPROCS - 1
            S_BND_PE(K) = ( ROWSX_PE(1, K+1) == 1 )
            E_BND_PE(K) = ( COLSX_PE(2, K+1) == GL_NCOLS )
            N_BND_PE(K) = ( ROWSX_PE(2, K+1) == GL_NROWS )
            W_BND_PE(K) = ( COLSX_PE(1, K+1) == 1 )

            GL_CZ(K) = COLSX_PE(1, K+1) - 1
            GL_RZ(K) = ROWSX_PE(1, K+1) - 1
            NC(K) = COLSX_PE(2, K+1) - GL_CZ(K)
            NR(K) = ROWSX_PE(2, K+1) - GL_RZ(K)
            NB(K) = 2 * (NC(K) + NR(K) + 2)

            SZ(K) = 0
            EZ(K) = NC(K) + 1
            NZ(K) = NC(K) + NR(K) + 3
            WZ(K) = 2 * NC(K) + NR(K) + 4

            GL_SZ(K) = GL_CZ(K)
            GL_EZ(K) = GL_RZ(K) + GL_NCOLS + 1
            GL_NZ(K) = GL_CZ(K) + GL_NCOLS + GL_NROWS + 3
            GL_WZ(K) = GL_RZ(K) + 2 * GL_NCOLS + GL_NROWS + 4
         END DO
      END IF                    ! FIRSTIME

      IERROR = 0
      LERROR = .FALSE.
      RERROR = .FALSE.

      IF ( MYPE == 0 ) THEN

         IF ( S_BND_PE(0) ) THEN
            IF ( E_BND_PE(0) ) THEN
               G_ARRAY(GL_SZ(0) + 1 : GL_SZ(0) + NC(0) + 1, :, :) =
     &              L_ARRAY(SZ(0) + 1 : SZ(0) + NC(0) + 1, :, :)
            ELSE
               G_ARRAY(GL_SZ(0) + 1 : GL_SZ(0) + NC(0), :, :) =
     &              L_ARRAY(SZ(0) + 1 : SZ(0) + NC(0), :, :)
            END IF
         END IF

         IF ( E_BND_PE(0) ) THEN
            IF ( N_BND_PE(0) ) THEN
               G_ARRAY(GL_EZ(0) + 1 : GL_EZ(0) + NR(0) + 1, :, :) =
     &              L_ARRAY(EZ(0) + 1 : EZ(0) + NR(0) + 1, :, :)
            ELSE
               G_ARRAY(GL_EZ(0) + 1 : GL_EZ(0) + NR(0), :, :) =
     &              L_ARRAY(EZ(0) + 1 : EZ(0) + NR(0), :, :)
            END IF
         END IF

         IF ( N_BND_PE(0) ) THEN
            IF ( W_BND_PE(0) ) THEN
               G_ARRAY(GL_NZ(0) : GL_NZ(0) + NC(0), :, :) =
     &              L_ARRAY(NZ(0) : NZ(0) + NC(0), :, :)
            ELSE
               G_ARRAY(GL_NZ(0) + 1 : GL_NZ(0) + NC(0), :, :) =
     &              L_ARRAY(NZ(0) + 1 : NZ(0) + NC(0), :, :)
            END IF
         END IF

         IF ( W_BND_PE(0) ) THEN
            IF ( S_BND_PE(0) ) THEN
               G_ARRAY(GL_WZ(0) : GL_WZ(0) + NR(0), :, :) =
     &              L_ARRAY(WZ(0) : WZ(0) + NR(0), :, :)
            ELSE
               G_ARRAY(GL_WZ(0) + 1 : GL_WZ(0) + NR(0), :, :) =
     &              L_ARRAY(WZ(0) + 1 : WZ(0) + NR(0), :, :)
            END IF
         END IF

C Processor 0 receives local arrays from all other processors and copies them
C to the global array; arrays are received in a first-come-first-serve order
         DO K = 1, NPROCS - 1
            CALL MPI_RECV(WHO, 1, MPI_INTEGER, MPI_ANY_SOURCE, TAG1,
     &           MPI_COMM_WORLD, ISTATUS, IERROR)
            IF ( IERROR /= 0 ) THEN
               XMSG = 'MPI error receiving processor id'
               CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
               LERROR = .TRUE.
            END IF

            ALLOCATE(
     &           RECV(1:NB(WHO), 1:NLAYS, 1:SIZE(G_ARRAY,3)),
     &           STAT = ALLOCSTAT)
            IF ( ALLOCSTAT /= 0 ) THEN
               XMSG = 'Failure allocating RECV'
               CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
            END IF

            CALL MPI_RECV(RECV, SIZE(RECV), MPI_REAL, WHO, TAG2,
     &           MPI_COMM_WORLD, ISTATUS, IERROR)
            IF ( IERROR /= 0 ) THEN
               XMSG = 'MPI error receiving data array'
               CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
               LERROR = .TRUE.
            END IF

            IF ( S_BND_PE(WHO) ) THEN
               IF ( E_BND_PE(WHO) ) THEN
                  G_ARRAY(GL_SZ(WHO) + 1 : GL_SZ(WHO) + NC(WHO) + 1, :, :) =
     &                 RECV(SZ(WHO) + 1 : SZ(WHO) + NC(WHO) + 1, :, :)
               ELSE
                  G_ARRAY(GL_SZ(WHO) + 1 : GL_SZ(WHO) + NC(WHO), :, :) =
     &                 RECV(SZ(WHO) + 1 : SZ(WHO) + NC(WHO), :, :)
               END IF
            END IF

            IF ( E_BND_PE(WHO) ) THEN
               IF ( N_BND_PE(WHO) ) THEN
                  G_ARRAY(GL_EZ(WHO) + 1 : GL_EZ(WHO) + NR(WHO) + 1, :, :) =
     &                 RECV(EZ(WHO) + 1 : EZ(WHO) + NR(WHO) + 1, :, :)
               ELSE
                  G_ARRAY(GL_EZ(WHO) + 1 : GL_EZ(WHO) + NR(WHO), :, :) =
     &                 RECV(EZ(WHO) + 1 : EZ(WHO) + NR(WHO), :, :)
               END IF
            END IF

            IF ( N_BND_PE(WHO) ) THEN
               IF ( W_BND_PE(WHO) ) THEN
                  G_ARRAY(GL_NZ(WHO) : GL_NZ(WHO) + NC(WHO), :, :) =
     &                 RECV(NZ(WHO) : NZ(WHO) + NC(WHO), :, :)
               ELSE
                  G_ARRAY(GL_NZ(WHO) + 1 : GL_NZ(WHO) + NC(WHO), :, :) =
     &                 RECV(NZ(WHO) + 1 : NZ(WHO) + NC(WHO), :, :)
               END IF
            END IF

            IF ( W_BND_PE(WHO) ) THEN
               IF ( S_BND_PE(WHO) ) THEN
                  G_ARRAY(GL_WZ(WHO) : GL_WZ(WHO) + NR(WHO), :, :) =
     &                 RECV(WZ(WHO) : WZ(WHO) + NR(WHO), :, :)
               ELSE
                  G_ARRAY(GL_WZ(WHO) + 1 : GL_WZ(WHO) + NR(WHO), :, :) =
     &                 RECV(WZ(WHO) + 1 : WZ(WHO) + NR(WHO), :, :)
               END IF
            END IF

            DEALLOCATE(RECV, STAT = ALLOCSTAT)
            IF ( ALLOCSTAT /= 0 ) THEN
               XMSG = 'Failure deallocating RECV'
               CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
            END IF
         END DO

      ELSE

C Each processor, except processor 0, sends its local array to processor 0
         WHO = MYPE

         CALL MPI_SEND(MYPE, 1, MPI_INTEGER, 0, TAG1,
     &        MPI_COMM_WORLD, IERROR)
         IF ( IERROR /= 0 ) THEN
            XMSG = 'MPI error sending processor id'
            CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
            LERROR = .TRUE.
         END IF

         CALL MPI_SEND(L_ARRAY, SIZE(L_ARRAY), MPI_REAL, 0, TAG2,
     &        MPI_COMM_WORLD, IERROR)
         IF ( IERROR /= 0 ) THEN
            XMSG = 'MPI error sending data array'
            CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
            LERROR = .TRUE.
         END IF

      END IF

C Notify all processors of failure and abort run
      CALL MPI_ALLREDUCE(LERROR, RERROR, 1, MPI_LOGICAL, MPI_LAND,
     &     MPI_COMM_WORLD, IERROR)
      IF ( IERROR /= 0 ) THEN
         XMSG = 'MPI_ALLREDUCE error'
         CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
         RERROR = .TRUE.
      END IF

      IF ( RERROR ) THEN
         XMSG = 'Failed to gather arrays on processor 0'
         CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT3)
      END IF

#else

      G_ARRAY(1:NBNDY, :, :) = L_ARRAY(1:NBNDY, :, :)

#endif

      END SUBROUTINE GATHER_BNDY_3D
