C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      SUBROUTINE GATHER_GRDD_4D(L_ARRAY, G_ARRAY, JDATE, JTIME)

C-----------------------------------------------------------------------
C Function:
C   Gather gridded 4D arrays from local processors in a global array on
C   processor 0

C Revision History:
C   Jul 2009 Peter Percell at UH-IMAQS: created
C-----------------------------------------------------------------------

      USE GRID_CONF             ! horizontal & vertical domain specifications

      IMPLICIT NONE

C Includes:
      INCLUDE SUBST_IOPARMS     ! I/O parameters definitions (XSTAT*)
#ifdef parallel
      INCLUDE SUBST_MPICH       ! MPI definitions and parameters
#endif

C Parameters:
      INTEGER, PARAMETER :: TAG1 = 901 ! MPI message tag for processor ID
      INTEGER, PARAMETER :: TAG2 = 902 ! MPI message tag for data array

C Arguments:
      REAL, INTENT(IN) :: L_ARRAY(:, :, :, :) ! local 4D array
      REAL, INTENT(OUT), TARGET :: G_ARRAY(:, :, :, :) ! global 4D array
      INTEGER, INTENT(IN) :: JDATE ! date, coded YYYYDDD
      INTEGER, INTENT(IN) :: JTIME ! time, coded HHMMSS

C Local Variables:
#ifdef parallel
      LOGICAL, SAVE :: FIRSTIME = .TRUE.

      CHARACTER(16) :: PNAME = 'GATHER_GRDD_4D'
      CHARACTER(96) :: XMSG = ' '

      INTEGER, SAVE :: NPROCS   ! total number of processors

      INTEGER :: K              ! loop counter

      INTEGER :: WHO            ! rank of sending processor
      INTEGER :: ISTATUS(MPI_STATUS_SIZE) ! MPI status code
      INTEGER :: IERROR         ! MPI error code
      LOGICAL :: LERROR         ! local error
      LOGICAL :: RERROR         ! local MPI ALLREDUCE error

      REAL, POINTER :: RECV(:, :, :, :)
#endif

C-----------------------------------------------------------------------

#ifdef parallel

      IF ( FIRSTIME ) THEN
         FIRSTIME = .FALSE.

C Get number of processors
         CALL MPI_COMM_SIZE(MPI_COMM_WORLD, NPROCS, IERROR)
         IF ( IERROR /= 0 ) THEN
            XMSG = 'Error in MPI_COMM_SIZE'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT3)
         END IF
      END IF

      IERROR = 0
      LERROR = .FALSE.
      RERROR = .FALSE.

      IF ( MYPE == 0 ) THEN

         G_ARRAY(1:NCOLS, 1:NROWS, :, :) = L_ARRAY(1:NCOLS, 1:NROWS, :, :)

C Processor 0 receives local arrays from all other processors and copies them
C to the global array; arrays are received in a first-come-first-serve order
         DO K = 1, NPROCS - 1
            CALL MPI_RECV(WHO, 1, MPI_INTEGER, MPI_ANY_SOURCE, TAG1,
     &           MPI_COMM_WORLD, ISTATUS, IERROR)
            IF ( IERROR /= 0 ) THEN
               XMSG = 'MPI error receiving processor id'
               CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
               LERROR = .TRUE.
            END IF

            RECV => G_ARRAY(
     &           COLSX_PE(1,WHO+1) : COLSX_PE(2,WHO+1),
     &           ROWSX_PE(1,WHO+1) : ROWSX_PE(2,WHO+1),
     &           :, :)

            CALL MPI_RECV(RECV, SIZE(RECV), MPI_REAL, WHO, TAG2,
     &           MPI_COMM_WORLD, ISTATUS, IERROR)
            IF ( IERROR /= 0 ) THEN
               XMSG = 'MPI error receiving data array'
               CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
               LERROR = .TRUE.
            END IF
         END DO

      ELSE

C Each processor, except processor 0, sends its local array to processor 0
         WHO = MYPE

         CALL MPI_SEND(MYPE, 1, MPI_INTEGER, 0, TAG1,
     &        MPI_COMM_WORLD, IERROR)
         IF ( IERROR /= 0 ) THEN
            XMSG = 'MPI error sending processor id'
            CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
            LERROR = .TRUE.
         END IF

         CALL MPI_SEND(L_ARRAY, SIZE(L_ARRAY), MPI_REAL, 0, TAG2,
     &        MPI_COMM_WORLD, IERROR)
         IF ( IERROR /= 0 ) THEN
            XMSG = 'MPI error sending data array'
            CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
            LERROR = .TRUE.
         END IF

      END IF

C Notify all processors of failure and abort run
      CALL MPI_ALLREDUCE(LERROR, RERROR, 1, MPI_LOGICAL, MPI_LAND,
     &     MPI_COMM_WORLD, IERROR)
      IF ( IERROR /= 0 ) THEN
         XMSG = 'MPI_ALLREDUCE error'
         CALL M3WARN(PNAME, JDATE, JTIME, XMSG)
         RERROR = .TRUE.
      END IF

      IF ( RERROR ) THEN
         XMSG = 'Failed to gather arrays on processor 0'
         CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT3)
      END IF

#else

      G_ARRAY(1:NCOLS, 1:NROWS, :, :) = L_ARRAY(1:NCOLS, 1:NROWS, :, :)

#endif

      END SUBROUTINE GATHER_GRDD_4D
