
C***********************************************************************
C   Portions of Models-3/CMAQ software were developed or based on      *
C   information from various groups: Federal Government employees,     *
C   contractors working on a United States Government contract, and    *
C   non-Federal sources (including research institutions).  These      *
C   research institutions have given the Government permission to      *
C   use, prepare derivative works, and distribute copies of their      *
C   work in Models-3/CMAQ to the public and to permit others to do     *
C   so.  EPA therefore grants similar permissions for use of the       *
C   Models-3/CMAQ software, but users are requested to provide copies  *
C   of derivative works to the Government without restrictions as to   *
C   use by others.  Users are responsible for acquiring their own      *
C   copies of commercial software associated with Models-3/CMAQ and    *
C   for complying with vendor requirements.  Software copyrights by    *
C   the MCNC Environmental Modeling Center are used with their         *
C   permissions subject to the above restrictions.                     *
C***********************************************************************

C RCS file, release, date & time of last delta, author, state, [and locker]
C $Header: /home/aqm/ppercell/aq/cmaq_4.6/models/CCTM/src/hadv/hppm/hppm.F,v 1.1.1.1 2006/09/22 18:19:30 sjr Exp $

C what(1) key, module and SID; SCCS file; date and time of last delta:
C %W% %P% %G% %U%

C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      SUBROUTINE HPPM(JDATE, JTIME, N, FIRST_CELL, LAST_CELL, ORI, DT, VEL, CON)

C----------------------------------------------------------------------
C Function:
C   This is the one-dimensional implementation of piecewise parabolic
C   method.  Variable grid spacing is allowed. The scheme is positive
C   definite and monotonic. It is conservative, and causes small
C   numerical diffusion.

C   A piecewise continuous parabola is used as the intepolation polynomial.
C   The slope of the parabola at cell edges are computed from a cumulative
C   function of the advected quantity.  These slopes are further modified
C   so that the interpolation function is monotone. For more detailed
C   information see:

C   Colella, P., and P. L. Woodward, (1984), "The Piecewise Parabolic
C   Method (PPM) for Gas-Dynamical Simulations," J. Comput. Phys. 54,
C   174-201.

C   The following definitions are used:

C              |---------------> Positive direction
C
C  -->|Boundary|<----------------Main Grid----------------->|Boundary|<--
C
C     |---><---|---><---|       ~|---><---|~       |---><---|---><---|
C       CON(0)   CON(1)            CON(i)            CON(n)  CON(n+1)
C
C     VEL(1)-->|        VEL(i)-->|        |-->VEL(i+1)      |-->VEL(n+1)
C
C      FP(0)-->|       FP(i-1)-->|        |-->FP(i)         |-->FP(n)
C
C      FM(1)<--|         FM(i)<--|        |<--FM(i+1)       |<--FM(n+1)
C
C                             -->| DS(i)  |<--

C Revision History:
C    04/30/07 by Peter Percell at UH-IMAQS:
C    - Rewrote for use with diagonal lines of E-grid cells
C----------------------------------------------------------------------

      USE HGRD_DEFN

      IMPLICIT NONE

C Includes:
      INCLUDE SUBST_IOPARMS     ! I/O parameters definitions

C Parameters:
#ifdef parallel
      INTEGER, PARAMETER :: XTHIK = 3
#else
      INTEGER, PARAMETER :: XTHIK = 1
#endif

#ifdef STEEPEN_FLAG
      LOGICAL, PARAMETER :: STEEPEN = .TRUE. ! Flag for discontinuty capturing (steepening)
#else
      LOGICAL, PARAMETER :: STEEPEN = .FALSE.
#endif

      REAL, PARAMETER :: ETA1 = 20.0
      REAL, PARAMETER :: ETA2 = 0.05
      REAL, PARAMETER :: EPS = 0.01

      REAL, PARAMETER :: TWO3RDS = 2.0 / 3.0

C Arguments:
      INTEGER, INTENT(IN)   :: JDATE ! current model date, coded YYYYDDD
      INTEGER, INTENT(IN)   :: JTIME ! current model time, coded HHMMSS
      INTEGER, INTENT(IN)   :: N ! number of unknown concentrations
      INTEGER, INTENT(IN)   :: FIRST_CELL ! first cell of extended line inside global boundary
      INTEGER, INTENT(IN)   :: LAST_CELL ! last cell of extended line inside global boundary
      CHARACTER, INTENT(IN) :: ORI ! orientation of advection ('C'-x or 'R'-y)
      REAL, INTENT(IN)      :: DT ! time step
      REAL, INTENT(IN)      :: VEL(1:) ! velocities at cell edges (flux points)
      REAL, INTENT(INOUT)   :: CON(1-XTHIK:, 1:)
                                ! concentrations along the line
                                !   if FIRST_CELL=1, CON(0,:) holds boundary values
                                !   if LAST_CELL=N, CON(N+1,:) holds boundary values

C Local Variables:
      CHARACTER, SAVE :: FIRSTORI = ' ' ! orientation flag (ORI)
      LOGICAL, SAVE :: FIRSTIME = .TRUE.

      CHARACTER(16) :: PNAME = 'HPPM'
      CHARACTER(96) :: XMSG = ' '

      INTEGER :: ALLOCSTAT

      INTEGER, SAVE :: NSPCS    ! number of species
      INTEGER :: MXN            ! maximum value of N

      REAL, ALLOCATABLE, SAVE :: DS(:) ! distance between cell edges

C Saved lattice variables for processor's extended domain
      REAL, ALLOCATABLE, SAVE :: CHI   (:) ! lattice var. for DC
      REAL, ALLOCATABLE, SAVE :: PSI   (:) ! lattice var. for DC
      REAL, ALLOCATABLE, SAVE :: MU    (:) ! lattice var. for CM
      REAL, ALLOCATABLE, SAVE :: NU    (:) ! lattice var. for CM
      REAL, ALLOCATABLE, SAVE :: LAMBDA(:) ! lattice var. for CM

      REAL, ALLOCATABLE, SAVE :: ZETA  (:) ! lattice var. for ETABAR
      REAL, ALLOCATABLE, SAVE :: SIGMA (:) ! lattice var. for D2C
      REAL, ALLOCATABLE, SAVE :: TAU   (:) ! lattice var. for D2C

C Temporary lattice variables
      REAL :: A
      REAL :: B
      REAL :: C
      REAL :: D
      REAL :: GAMMA

C Temporary lattice variable arrays
      REAL :: ALPHA(2-XTHIK : N+XTHIK-1)
      REAL :: BETA (2-XTHIK : N+XTHIK-1)

      REAL :: FM(1:N+1, 1:SIZE(CON, 2)) ! outflux from left or bottom of cell
      REAL :: FP(0:N, 1:SIZE(CON, 2)) ! outflux from right or top of cell

! zone R.H. trial intercept
      REAL :: CM(MAX(FIRST_CELL,0)   : MIN(LAST_CELL,N+1)+1, 1:SIZE(CON, 2))
! CR - CL
      REAL :: DC(MAX(FIRST_CELL,0)-1 : MIN(LAST_CELL,N+1)+1, 1:SIZE(CON, 2))
! second derivative
      REAL :: D2C(FIRST_CELL+1       : LAST_CELL-1,          1:SIZE(CON, 2))
! zone L.H. intercept
      REAL :: CL(MAX(FIRST_CELL,0)   : MIN(LAST_CELL,N+1),   1:SIZE(CON, 2))
! zone R.H. intercept
      REAL :: CR(MAX(FIRST_CELL,0)   : MIN(LAST_CELL,N+1),   1:SIZE(CON, 2))
! coefficient of second-order term
      REAL :: C6(MAX(FIRST_CELL,0)   : MIN(LAST_CELL,N+1),   1:SIZE(CON, 2))
! discontinuity homotopy function
      REAL :: ETA(MAX(FIRST_CELL,0)  : MIN(LAST_CELL,N+1),   1:SIZE(CON, 2))
! zone L.H. intercept w/ discontinuity
      REAL :: CLD(MAX(FIRST_CELL,0)  : MIN(LAST_CELL,N+1),   1:SIZE(CON, 2))
! zone R.H. intercept w/ discontinuity
      REAL :: CRD(MAX(FIRST_CELL,0)  : MIN(LAST_CELL,N+1),   1:SIZE(CON, 2))

      REAL :: ETABAR            ! 3rd to 1st order derivative ratio

      REAL :: X                 ! Courant number
      REAL :: Y

      INTEGER :: I              ! loop index
      INTEGER :: S              ! species index

      INTEGER, SAVE :: FIRST_PAR ! first cell with parabolic concentration distribution
      INTEGER, SAVE :: LAST_PAR ! last cell with parabolic concentration distribution

      INTEGER, SAVE :: FIRST_STD ! first cell with standard parabolic distribution
      INTEGER, SAVE :: LAST_STD ! last cell with standard parabolic distribution

C----------------------------------------------------------------------

      IF ( FIRSTIME ) THEN
         FIRSTIME = .FALSE.
         NSPCS = SIZE(CON, 2)
         MXN = MAX(NCOLS, NROWS)

         ALLOCATE(
     &        DS    (1-XTHIK : MXN+XTHIK),
     &        CHI   (2-XTHIK : MXN+XTHIK-1),
     &        PSI   (2-XTHIK : MXN+XTHIK-1),
     &        MU    (3-XTHIK : MXN+XTHIK-1),
     &        NU    (3-XTHIK : MXN+XTHIK-1),
     &        LAMBDA(3-XTHIK : MXN+XTHIK-1),
     &        STAT = ALLOCSTAT)
         IF ( ALLOCSTAT /= 0 ) THEN
            XMSG = 'Failure allocating lattice arrays'
            CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
         END IF

         IF ( STEEPEN ) THEN
            ALLOCATE(
     &           ZETA (3-XTHIK : MXN+XTHIK-2),
     &           SIGMA(2-XTHIK : MXN+XTHIK-1),
     &           TAU  (2-XTHIK : MXN+XTHIK-1),
     &           STAT = ALLOCSTAT)
            IF ( ALLOCSTAT /= 0 ) THEN
               XMSG = 'Failure allocating steepening lattice arrays'
               CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
            END IF
         END IF
      END IF                    ! FIRSTIME

      IF ( ORI /= FIRSTORI ) THEN
         FIRSTORI = ORI

         IF ( ORI == 'C' ) THEN
            DS(:) = XCELL_GD
         ELSE
            DS(:) = YCELL_GD
         END IF

C Cells inside the global domain are given a parabolic concentration profile,
C but cells near the global boundary require special handling because some of
C the concentration data needed for typical handling is not available; cells
C in the global boundary are given a uniform concentration profile
         FIRST_PAR = MAX(FIRST_CELL, 0)
         LAST_PAR = MIN(LAST_CELL, N+1)

         FIRST_STD = FIRST_CELL + 2
         LAST_STD = LAST_CELL - 2

         DO I = FIRST_STD - 1, LAST_STD + 1
            ALPHA( I ) = DS( I )   + DS( I+1 )
            BETA( I )  = DS( I-1 ) + DS( I )
            D = DS( I ) / ( BETA( I ) + DS( I+1 ) )
            CHI( I ) = D * ( DS( I-1 ) + BETA( I ) ) / ALPHA( I )
            PSI( I ) = D * ( ALPHA( I ) + DS( I+1 ) ) / BETA( I )
         END DO

         DO I = FIRST_STD, LAST_STD + 1
            GAMMA = DS( I-2 ) + DS( I-1 )
            A = DS( I-1 ) / BETA( I )
            B = 2.0 * DS( I ) / BETA( I )
            C = 1.0 / ( ALPHA( I ) + GAMMA )
            MU( I ) = C * DS( I-1 ) * GAMMA / ( DS( I-1 ) + BETA( I ) )
            NU( I ) = C * DS( I ) * ALPHA( I ) / ( DS( I ) + BETA( I ) )
            LAMBDA( I ) = A + MU( I ) * B - 2.0 * NU( I ) * A
         END DO

         IF ( STEEPEN ) THEN
            DO I = FIRST_STD, LAST_STD
               ZETA( I ) = 0.25 * ( ALPHA( I ) * ALPHA( I )
     &              - ALPHA( I ) * BETA( I )
     &              + BETA( I ) * BETA( I ) )
            END DO

            DO I = FIRST_STD - 1, LAST_STD + 1
               C = 1.0 / ( BETA( I ) + DS( I+1 ) )
               SIGMA( I ) = C / ALPHA( I )
               TAU( I )   = C / BETA( I )
            END DO
         END IF
      END IF                    ! FIRSTORI

C Set all fluxes to zero. Either positive or negative flux will
C remain zero depending on the sign of the velocity.
      FM = 0.0
      FP = 0.0

C Second order polynomial inside the domain
      DO S = 1, NSPCS
         DO I = FIRST_STD - 1, LAST_STD + 1

C Compute average slope in the i'th zone

C Equation (1.7)
            DC( I,S ) = CHI( I ) * ( CON( I+1,S ) - CON( I,S ) )
     &           + PSI( I ) * ( CON( I,S )   - CON( I-1,S ) )

C Guarantee that CM lies between CON(I) and CON(I+1) - monotonicity constraint
            IF ( ( CON( I+1,S ) - CON( I,S ) )
     &           * ( CON( I,S )   - CON( I-1,S ) ) .GT. 0.0 ) THEN
               DC( I,S ) = SIGN( 1.0, DC( I,S ) )
     &              * MIN(      ABS( DC( I,S ) ),
     &              2.0 * ABS( CON( I+1,S ) - CON( I,S ) ),
     &              2.0 * ABS( CON( I,S ) - CON( I-1,S ) ) )
            ELSE
               DC( I,S ) = 0.0
            END IF              ! Equation (1.8)

         END DO                 ! I
      END DO                    ! S

C If the line of cells starts in the global boundary...
C Zeroth order polynomial at the boundary cells
C First order polynomial at the next cells, no monotonicity constraint needed
      IF ( FIRST_CELL == 1 ) THEN
         I = FIRST_STD - 2
         DO S = 1, NSPCS
            IF ( I >= FIRST_PAR ) CM( I,S ) = CON( I,S )
            CM( I+1,S ) = ( DS( I ) * CON( I+1,S )
     &           +   DS( I+1 ) * CON( I,S ) )
     &           / ( DS( I ) + DS( I+1 ) )
         END DO
      END IF

C If the line of cells ends in the global boundary...
C Zeroth order polynomial at the boundary cells
C First order polynomial at the next cells, no monotonicity constraint needed
      IF ( LAST_CELL == N ) THEN
         I = LAST_STD + 2
         DO S = 1, NSPCS
            IF ( I <= LAST_PAR ) CM( I+1,S ) = CON( I,S )
            CM( I,S ) = ( DS( I-1 ) * CON( I,S )
     &           +   DS( I )   * CON( I-1,S ) )
     &           / ( DS( I-1 ) + DS( I ) )
         END DO
      END IF

      DO S = 1, NSPCS
         DO I = FIRST_STD, LAST_STD + 1 ! Equation (1.6)
            CM( I,S ) = CON( I-1,S )
     &           + LAMBDA( I ) * ( CON( I,S ) - CON( I-1,S ) )
     &           - MU( I ) * DC( I,S ) + NU( I ) * DC( I-1,S )
         END DO

C Initialize variables for discontinuty capturing. This is necessary
C even if discontinuity capturing is deactivated.
         DO I = FIRST_PAR, LAST_PAR ! these reduce Eqn. 1.15
            ETA( I,S ) = 0.0
            CLD( I,S ) = CON( I,S )
            CRD( I,S ) = CON( I,S )
         END DO
      END DO                    ! S

      IF ( STEEPEN ) THEN

C Finite diff. approximation to 2nd derivative as in Equation (1.17)
         DO S = 1, NSPCS
            DO I = FIRST_STD - 1, LAST_STD + 1
               D2C( I,S ) = SIGMA( I ) * ( CON( I+1,S ) - CON( I,S ) )
     &              - TAU( I ) * ( CON( I,S ) - CON( I-1,S ) )
            END DO
         END DO

C No discontinuity detection near the boundary: cells 1, 2, N-1, N
         DO S = 1, NSPCS
            DO I = FIRST_STD, LAST_STD

C Compute etabars in Equation (1.16)
               IF ( ( - D2C( I+1,S ) * D2C( I-1,S ) .GT. 0.0 ) .AND.
     &              (
     &              ABS( CON( I+1,S ) - CON( I-1,S ) )
     &              - EPS * MIN( ABS( CON( I+1,S ) ), ABS( CON( I-1,S ) )
     &              )
     &              .GT. 0.0 ) ) THEN ! 2nd derivative changes sign
                  ETABAR = - ZETA( I )
     &                 * ( D2C( I+1,S ) - D2C( I-1,S ) )
     &                 / ( CON( I+1,S ) - CON( I-1,S ) )
               ELSE
                  ETABAR = 0.0
               END IF

C Equation (1.16)
               ETA( I,S ) = MAX( 0.0,
     &              MIN( ETA1 * ( ETABAR - ETA2 ), 1.0 ) )

C Equation (1.14)
               CRD( I,S ) = CON( I+1,S ) - 0.5 * DC( I+1,S )
               CLD( I,S ) = CON( I-1,S ) + 0.5 * DC( I-1,S )

            END DO              ! I
         END DO                 ! S

      END IF                    ! STEEPEN

C Generate piecewise parabolic distributions
      DO S = 1, NSPCS

         DO I = FIRST_PAR, LAST_PAR

C Equation (1.15)
            CR( I,S ) = CM( I+1,S )
     &           + ETA( I,S ) * ( CRD( I,S ) - CM( I+1,S ) )
            CL( I,S ) = CM( I,S )
     &           + ETA( I,S ) * ( CLD( I,S ) - CM( I,S ) )

C Monotonicity
            IF ( ( CR( I,S ) - CON( I,S ) )
     &           * ( CON( I,S ) - CL( I,S ) ) .GT. 0.0 ) THEN

C Temporary computation of DC and C6
               DC( I,S ) = CR( I,S ) - CL( I,S )
               C6( I,S ) = 6.0 * ( CON( I,S )
     &              - 0.5 * ( CL( I,S ) + CR( I,S ) ) )

C Overshoot cases
               IF ( DC( I,S ) * C6( I,S ) .GT.
     &              DC( I,S ) * DC( I,S ) ) THEN
                  CL( I,S ) = 3.0 * CON( I,S ) - 2.0 * CR( I,S )
               ELSE IF ( -DC( I,S ) * DC( I,S ) .GT.
     &                 DC( I,S ) * C6( I,S ) ) THEN
                  CR( I,S ) = 3.0 * CON( I,S ) - 2.0 * CL( I,S )
               END IF

            ELSE                ! Local extremum: Interpolation
                                ! function is set to be a constant
               CL( I,S ) = CON( I,S )
               CR( I,S ) = CON( I,S )

            END IF

            DC( I,S ) = CR( I,S ) - CL( I,S ) ! Equation (1.5)
            C6( I,S ) = 6.0 * ( CON( I,S )
     &           - 0.5 * ( CL( I,S ) + CR( I,S ) ) )

         END DO                 ! I

C Compute fluxes from cells inside the global boundary using the parabolic
C distribution as in Equation (1.12)

C mass leaving interval at an end
C    = length of segment leaving * average concentration in that segment
         IF ( FIRST_PAR <= 0 .AND. VEL( 1 ) .GT. 0.0 ) THEN
            Y = VEL( 1 ) * DT
            X = Y / DS( 0 )
            FP( 0,S ) = Y * ( CR( 0,S ) - 0.5 * X * ( DC( 0,S )
     &           - C6( 0,S ) * ( 1.0 - TWO3RDS * X ) ) )
         END IF

         DO I = 1, N
            IF ( VEL( I ) .LT. 0.0 ) THEN ! mass leaving at left end
               Y = -VEL( I ) * DT
               X = Y / DS( I )
               FM( I,S ) = Y * ( CL( I,S ) + 0.5 * X * ( DC( I,S )
     &              + C6( I,S ) * ( 1.0 - TWO3RDS * X ) ) )
            END IF

            IF ( VEL( I+1 ) .GT. 0.0 ) THEN ! mass leaving at right end
               Y = VEL( I+1 ) * DT
               X = Y / DS( I )
               FP( I,S ) = Y * ( CR( I,S ) - 0.5 * X * ( DC( I,S )
     &              - C6( I,S ) * ( 1.0 - TWO3RDS * X ) ) )
            END IF
         END DO

         I = N + 1
         IF ( LAST_PAR >= I .AND. VEL( I ) .LT. 0.0 ) THEN
            Y = -VEL( I ) * DT
            X = Y / DS( I )
            FM( I,S ) = Y * ( CL( I,S ) + 0.5 * X * ( DC( I,S )
     &           + C6( I,S ) * ( 1.0 - TWO3RDS * X ) ) )
         END IF

      END DO                    ! S

C Compute fluxes from boundary cells assuming uniform distribution
      IF ( FIRST_PAR == 1 ) THEN
         IF ( VEL( 1 ) .GT. 0.0 ) THEN
            Y = VEL( 1 ) * DT
            DO S = 1, NSPCS
               FP( 0,S ) = Y * CON( 0,S )
            END DO
         END IF
      END IF

      IF ( LAST_PAR == N ) THEN
         I = N + 1
         IF ( VEL( I ) .LT. 0.0 ) THEN
            Y = -VEL( I ) * DT
            DO S = 1, NSPCS
               FM( I,S ) = Y * CON( I,S )
            END DO
         END IF
      END IF

C Update concentrations as in Equation (1.13)
      DO S = 1, NSPCS
         DO I = 1, N
            CON( I,S ) = CON( I,S )
     &           + ( FP( I-1,S ) - FP( I,S )
     &           +   FM( I+1,S ) - FM( I,S ) ) / DS( I )
         END DO
      END DO

      RETURN
      END
