
C***********************************************************************
C   Portions of Models-3/CMAQ software were developed or based on      *
C   information from various groups: Federal Government employees,     *
C   contractors working on a United States Government contract, and    *
C   non-Federal sources (including research institutions).  These      *
C   research institutions have given the Government permission to      *
C   use, prepare derivative works, and distribute copies of their      *
C   work in Models-3/CMAQ to the public and to permit others to do     *
C   so.  EPA therefore grants similar permissions for use of the       *
C   Models-3/CMAQ software, but users are requested to provide copies  *
C   of derivative works to the Government without restrictions as to   *
C   use by others.  Users are responsible for acquiring their own      *
C   copies of commercial software associated with Models-3/CMAQ and    *
C   for complying with vendor requirements.  Software copyrights by    *
C   the MCNC Environmental Modeling Center are used with their         *
C   permissions subject to the above restrictions.                     *
C***********************************************************************

C:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
      PROGRAM DRIVER

C-----------------------------------------------------------------------
C Function:
C   Driver for sensitivity analysis

C Revision History:
C   Apr 2007 Kumaresh Singh and Adrian Sandu at Virginia Tech: created
C   Aug 2007 Peter Percell at UH-IMAQS: rewrote
C-----------------------------------------------------------------------

      USE PCGRID_DEFN           ! inherits HGRD_DEFN and CGRID_SPCS
      USE VGRD_DEFN             ! vertical layer specifications
      USE CGRID_SPCS            ! CGRID species number and offsets
      USE ADJOINT_FILES         ! adjoint file data and functions

      IMPLICIT NONE

C Include Files:
      INCLUDE SUBST_IOPARMS     ! I/O parameters definitions
      INCLUDE SUBST_IODECL      ! I/O definitions and declarations
      INCLUDE SUBST_FILES_ID    ! file name parameters

C Load the mechanism COMMON ... (next two INCLUDE files)
      INCLUDE SUBST_RXCMMN      ! chemical mechamism reactions COMMON
      INCLUDE SUBST_RXDATA      ! chemical mechamism reactions DATA

C Local Variables:
      CHARACTER(16) :: PNAME = 'DRIVER'
      CHARACTER(96) :: XMSG = ' '

      REAL :: CLOCK             ! Wall-clock time (sec) at initialization.
      INTEGER :: PAR_ERR        ! Error code from parallel initialization.

      INTEGER :: SDATE          ! start date, format YYYYDDD
      INTEGER :: STIME          ! start time, format HHMMSS
      INTEGER :: TSTEP(2)       ! time step vector (HHMMSS)
                                ! TSTEP(1) = local output step
                                ! TSTEP(2) = sciproc sync. step (chem)
      INTEGER :: NSTEPS         ! run duration: number of output time steps
      INTEGER :: NREPS          ! number of sync steps per output step

      INTEGER :: JDATE          ! current model date, coded YYYYDDD
      INTEGER :: JTIME          ! current model time, coded HHMMSS

      REAL, POINTER :: CGRID(:,:,:,:)

      INTEGER :: DAY_NUM        ! day number

      LOGICAL :: XFIRST

      INTERFACE
         SUBROUTINE INITSCEN(CGRID, SDATE, STIME, TSTEP, NSTEPS)
            IMPLICIT NONE
            REAL, POINTER        :: CGRID(:, :, :, :)
            INTEGER, INTENT(OUT) :: SDATE, STIME
            INTEGER, INTENT(OUT) :: TSTEP(2)
            INTEGER, INTENT(OUT) :: NSTEPS
         END SUBROUTINE INITSCEN
         SUBROUTINE DRIVER_FWD(CGRID, JDATE, JTIME, TSTEP, NSTEPS, NREPS,
     &     DAY_NUM, XFIRST)
            IMPLICIT NONE
            REAL, POINTER          :: CGRID(:,:,:,:)
            INTEGER, INTENT(INOUT) :: JDATE, JTIME
            INTEGER, INTENT(IN)    :: TSTEP(2), NSTEPS
            INTEGER, INTENT(IN)    :: NREPS
            INTEGER, INTENT(OUT)   :: DAY_NUM
            LOGICAL, INTENT(INOUT) :: XFIRST
         END SUBROUTINE DRIVER_FWD
         SUBROUTINE DRIVER_BWD(CGRID, JDATE, JTIME, TSTEP, NSTEPS, NREPS,
     &     DAY_NUM, XFIRST)
            IMPLICIT NONE
            REAL, POINTER          :: CGRID(:,:,:,:)
            INTEGER, INTENT(INOUT) :: JDATE, JTIME
            INTEGER, INTENT(IN)    :: TSTEP(2), NSTEPS
            INTEGER, INTENT(IN)    :: NREPS
            INTEGER, INTENT(INOUT) :: DAY_NUM
            LOGICAL, INTENT(INOUT) :: XFIRST
         END SUBROUTINE DRIVER_BWD
      END INTERFACE

C-----------------------------------------------------------------------

C Get number of species, and starting indices for CGRID array.
      CALL CGRID_MAP(NSPCSD, GC_STRT, AE_STRT, NR_STRT, TR_STRT)

C Start up processor communication and define horizontal domain decomposition
C and vertical layer structure
      CALL PAR_INIT(PNAME, NSPCSD, CLOCK, PAR_ERR)
      IF ( PAR_ERR /= 0 ) THEN
         XMSG = 'Error in PAR_INIT'
         CALL M3EXIT(PNAME, 0, 0, XMSG, XSTAT2)
      END IF

C Initialize PCGRID
      IF ( .NOT. PCGRID_INIT() ) THEN
         XMSG = 'Failure defining horizontal domain'
         CALL M3EXIT(PNAME, 0, 0, XMSG, XSTAT2)
      END IF

      CGRID => PCGRID(1:NCOLS, 1:NROWS, :, :)

C Initialize conc field: Copy IC's to CONC file as step 0
C Convention: the input file concentration units are always ppmV.
      CALL INITSCEN(CGRID, SDATE, STIME, TSTEP, NSTEPS)

C Verify input file header consistency with COORD.EXT and run duration
      CALL FLCHECK(SDATE, STIME, TSTEP(1))

C Set the current date and time to the start date and time
      JDATE = SDATE
      JTIME = STIME

C Get the synchronization step, TSTEP(2), and number of repetitions per output
C step, NREPS
C (Note: TSTEP(2) and NREPS are constant for a run because checkpoints are
C written every sync step and the IOAPI requires a constant output time step.)
      CALL SYNCSTEP(JDATE, JTIME, TSTEP, NREPS)

C Create new adjoint files and open for update
      CALL OP_ADJ_FILE(JDATE, JTIME, TSTEP, ADJ_CHEM_CHK)

      CALL OP_ADJ_FILE(JDATE, JTIME, TSTEP, ADJ_LGRID)
      CALL OP_ADJ_FILE(JDATE, JTIME, TSTEP, ADJ_EMGRID)

C Do forward simulation and write checkpoint states
      XFIRST = .TRUE.
      CALL DRIVER_FWD(CGRID, JDATE, JTIME, TSTEP, NSTEPS, NREPS,
     &     DAY_NUM, XFIRST)

C Open checkpoint files as read-only
      IF ( .NOT. OPEN3(ADJ_CHEM_CHK, FSREAD3, PNAME) ) THEN
         XMSG = 'Could not open ' // TRIM(ADJ_CHEM_CHK) //
     &        ' file as OLD:READ-ONLY'
         CALL M3EXIT(PNAME, JDATE, JTIME, XMSG, XSTAT1)
      END IF

C Do backward (adjoint) solution and write sensitivities
      CALL DRIVER_BWD(CGRID, JDATE, JTIME, TSTEP, NSTEPS, NREPS,
     &     DAY_NUM, XFIRST)

C Clean up communications
      CALL PAR_TERM(PNAME, JDATE, JTIME, CLOCK)

      END PROGRAM DRIVER
